###############################################################################
# IMPORTS 
#from its import *
from Numeric import *
from Matrix import Matrix
from UserArray import *
from copy import *
import types

import Tkinter,Pmw,gui
import tkFileDialog
#from gui import *
#from classlib import gui
#from  gui import Panel

###############################################################################

#
###############################################################################

class Costs_handling:
    """
    This class provides mix in methods to handle costs.
    These general methods are supposed to be used in both
    simulation objects and modules, but need to be overwritten.

    """
    
        
    def init_costs(self):
        """
        Initializes esimated initial costs and costs per year.

        Should be overwritten by objects to set initial estimated costs.
        """
        self.set_costs()
        self.reset_costs()
        
    def set_costs(self,costs=None,lifetime_estim=-1):
        """
        Sets multiple costs, and optionally
        lifetime estimate.
        """
        if costs:
            for label in costs.keys():
                if self.costs.has_key(label):
                    self.set_cost(label,costs[label])
                else:
                    self.add_cost_item(label,costs[label])
            
        if lifetime_estim>0:
            self.set_lifetime_estim(lifetime_estim)
        else:
            self.set_lifetime_estim(-1)

        #if quantity:
        #    self.set_quantity(quantity)
        #else:
        #    self.quantity=0
        #    self.set_quantity(self.get_quantity(auto_flag=1))

    def set_cost(self,label,cost):
        """
        ONLY function to change attributes and values of a cost item.
        """
        if self.costs.has_key(label):
            # this is a lazy method, should hallen in a more controlled way.
            self.costs[label].update(cost)
        else:
            print 'WARNING: Cost label "',label,'" in ',self.id,'not available'

    def set_cost_value(self,label,value):
        """
        Change VALUE of a cost item.
        """
        self.set_cost(label,{'value':value})

    def set_cost_table(self,label,table):
        """
        Change cost table of a cost item.
        """
        if self.costs[label].has_key('table'):
            self.costs[label]['table']=table

    def add_cost_value(self,label,value):
        """
        Add costs to the labeled cost item.
        This is only possible if costs is a scalar value.
        """
        if self.costs[label].has_key('value'):
            self.set_cost_value(label,self.costs[label]['value']+value)
            
    def get_cost_value(self,label,quantity=None):
        """
        Function to return value. If cost is a table, the returned
        value depends on the quantity.
        """
        cost=self.costs[label]
        if cost.has_key('value'):
            return cost['value']
        
        elif cost.has_key('function'):
            return self.cost_functions[cost['function']](label,quantity)
            
        elif cost.has_key('table'):
            if not quantity:
                quantity=1
            
            # get cost for smallest quantity
            q_min=cost['table'].keys()[0]
            c_max=cost['table'][q_min]

            for q in cost['table'].keys():
                c=cost['table'][q]
                if q<q_min:
                    q_min=q
                    c_max=c
            
            for q in cost['table'].keys():
                c=cost['table'][q]
                if (quantity>=q)&(c<c_max):
                    c_max=c
                
            return c_max
        
    def add_cost_item(self,label,cost):
        """
        Add a cost item, with a label to costs directory.
        """
        self.costs[label]={'mode':'read',
                           'currency':self.sim.currency}
        self.set_cost(label,cost)

    def del_cost_item(self,label):
        """
        Delete a cost item, with a label from costs directory.
        """
        if self.costs.has_key(label):
            del self.costs[label]
        else:
            print 'WARNING: there is no',label,'in',self.id,'costs.'
        
        
    def reset_costs(self):
        """
        Sets all costs in the costs directory that occurred during
        the current simulation to zero.

        To be overwritten.
        """
        pass
        
    
    def do_costs_current(self):
        """
        Calculates current costs and modifies the costs dictionary.
                
        """
        pass

    def get_cost_descriptor(self,label):
        """
        Returns formatting instructions for cost data of cost item with label.
        """
        
        #if self.table:
        #    cost_max=max(self.table.values())
        #else:
        #    cost_max=self.value
        cost=self.costs[label]
        descriptor={'name':cost['name'],
                    'key':label,
                    'type':'real',
                    'printformat':'%9.2f',
                    'width':1,
                    'mode':cost['mode'],
                    'valid':{'min':0,'max':10**8},
                    'units':{cost['currency']:1},
                    }

        # add button if there is a quantity cost table
        if cost.has_key('table'):
            if cost['mode']=='write':
                descriptor['mode']='read'
                descriptor['button']={'text':'Edit...',
                                      'command':self.edit_costtable_gui}
        return descriptor

    def edit_costtable_gui(self,costsdatawidget,descriptor,value):
        """
        Creating widget to edit cost-table.
        """
        # set descriptor d for edit widget to write mode
        #print 'edit_gui',self.name
        d=descriptor.copy()
        d['mode']='write'
        cost=self.costs[descriptor['key']]
        gui.CostEdit(
            d,
            cost['table'],
            cost['currency'],costsdatawidget,
            applybutton_command=self.modify_cost_gui)

            
    def modify_cost_gui(self,descriptor,coststable,costsdatawidget):
        """
        Make changes to cost table using edit_gui 
        """
        #print 'modify_costs:'
        #print ' ',descriptor['key'],'=',coststable
        self.set_cost_table(descriptor['key'],coststable)
        descriptors,dataset=self.format_costs()
        costsdatawidget.update_contents(dataset)
        #print '  self.table=',self.table
        #self.costedit_gui=None

   
    def set_lifetime_estim(self,lifetime_estim):
        """
        Return estimated lifetime in years.
        """

        self.lifetime_estim=lifetime_estim

    
    def get_lifetime_estim(self):
        """
        Return estimated lifetime in years.
        """

        return self.lifetime_estim
    
    #def set_quantity(self,quantity=0):
    #    self.quantity=quantity

    def get_quantity(self):
        """
        Defaults to one
        or the number of modules in case of a module containing
        simulation object.
        """
        if hasattr(self,'obj'):
            if hasattr(self.obj,'mods'):
                quantity=len(self.obj.mods)
            else:
                quantity=1
        else:
            quantity=1

        return quantity
    

    def make_costspage(self,parent,widgets):
        """
        Default cost gui widget creation.
        Creates widgets in parent frame. All widgets are put in the
        widgets directory with a string as keyword. These widgets
        can be configured in the corresponding update method.

        May be overwritten.
        """
        descriptor,dataset=self.format_costs()
        widgets['costsdata']=gui.DatasetEditor(descriptor,dataset,
                                               parent,
                                               applybutton_command=\
                                                   self.on_apply_costs,
                                               applyentry_command=\
                                                   self.on_apply_costs,
                                               )
       
        
        
        
        self.make_costspage_specific(parent,widgets)
        widgets['costsdata'].pack(expand=1,side = 'top',fill='both' )

    def make_costspage_specific(self,parent,widgets):
        """
        Object/Module specific cost page.
        """
        pass
    

    def on_apply_costs(self,costsdatawidget=None,costsdata=None):
        """
        Apply cost data on cost page to this module.
        """
        #print 'on_apply_costs',self.id
        
        costsdata=costsdatawidget.get_contents()
        # set all cost values from widget
        for label in self.costs.keys():
            # if item has no table, in which case the table editior did it 
            if not self.costs[label].has_key('table'):
                self.set_cost_value(label,costsdata[label])

                           
        
        if costsdata.has_key('Estimated life-time'):
            self.set_lifetime_estim(costsdata['Estimated life-time'])
        descriptor,dataset=self.format_costs()
        costsdatawidget.update_contents(dataset)
        
    def on_apply_costs_to_all(self,costsdatawidget=None):
        """
        Apply cost data on cost page to all modules of the same type
        and with the same track management.
        """
        #print 'on_apply_costs_to_all',self.id
        self.on_apply_costs(costsdatawidget)
        for mod in self.obj.mods.values():
            if mod.man_id==self.man_id:
                mod.set_costs(self.costs,self.get_lifetime_estim())
        
    def update_costspage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        self.do_costs_current()
        descriptor,dataset=self.format_costs()
        widgets['costsdata'].update_contents(dataset)

    def format_costs(self):
        """
        Return formatted costdata.
        """
        # update object internal cost calcs
        self.do_costs_current()
        
        quantity=self.get_quantity()
        # this is a general algorithm that should give reasonable results
        # if this does nor work: overwrite
        cost_labels=self.costs.keys()
        cost_labels.sort()
        descriptors=[]
        table={}
        for label in cost_labels:
            #print 'format',self.costs[label].name,self.costs[label][quantity]
            descriptors.append(self.get_cost_descriptor(label))
            table[label]=self.get_cost_value(label,quantity)

        descriptors.append({'name':'Estimated life-time',
                            'key':'lifetime_estim', 
                            'type':'integer',
                            'valid':{'min':0,'max':10**8},
                            'width':1,
                            'mode':'write',
                            'units':{'a':1}}
                           )
        
        table['lifetime_estim']=self.get_lifetime_estim()
        
        self.format_costs_specific(quantity,descriptors,table)
        return descriptors,table
    
    def format_costs_specific(self,quantity,descriptors,table):
        """
        Object/Module specific cost formatting.
        """
        pass
       
  
    #def on_reset_quantity(self,costsdatawidget,descriptor,value):
    #    #entrywidget.setentry(str(self.get_quantity()))
    #    #self.set_quantity(self.get_quantity())
    #    descriptor,dataset=self.format_costs(self.get_quantity(auto_flag=1))
    #    costsdatawidget.update_contents(dataset)

    def on_costsdata_entry(self,costsdatawidget,costsdata):
        #print 'on_costsdata_entry',costsdata['Quantity']
        #if costsdata.has_key('Quantity'):
        #    descriptor,dataset=self.format_costs(costsdata['Quantity'])
        #else:
        descriptor,dataset=self.format_costs()
        costsdatawidget.update_contents(dataset)
        

class Cost:
    """
    Basic cost class.
    """
    def __init__(self,obj,name,value=0,table=None,currency=None,mode='read'):
        self.obj=obj
        self.name=name
        self.value=value
        if currency:
            self.currency=currency
        elif obj:
            self.currency=self.obj.sim.currency
            
        self.table=table
        self.mode=mode

    def has_costtable(self):
        if self.table: return 1
        else: return 0

    def get_descriptor(self):
        """
        Returns formatting instructions for cost data.
        """
        
        #if self.table:
        #    cost_max=max(self.table.values())
        #else:
        #    cost_max=self.value
            
        descriptor={'key':self.name,
                    'type':'real',
                    'printformat':'%.2f',
                    'width':1,
                    'mode':self.mode,
                    'valid':{'min':0,'max':10**8},
                    'units':{self.currency:1},
                    }

        # add button if there is a quantity cost table
        if self.table:
            if self.mode=='write':
                descriptor['button']={'text':'Edit...',
                                      'command':self.edit_gui}
        return descriptor
    
    def format(self,quantity=None):
        """
        Returns descriptor information and current value for a given
        quantity. If no quantity is given the set quantity is used.
        """
        
        return self.get_descriptor(), self.__getitem__[quantity]
    
    def __getitem__(self,quantity=None):

        return self.get_value(quantity)
    
    def add_cost(self,cost,quantity=1):
        if self.currency == cost.currency:
            self.set_value(self.value+cost[quantity])
        else:
            print 'ERROR: could not convert currencies'

    def add(self,cost,currency=None):
        if currency:
            value=cost
        else:
            if len(cost)==2:
                value,currency=cost
                
        if self.currency == currency:
            self.value+=value
        else:
            print 'ERROR: could not convert currencies'
        
    def get_value(self,quantity=None):
        """
        Central function to return value. If these is a table, the returned
        value depends on the quantity.
        """
        if not self.table:
            return self.value
        else:
            if not quantity:
                quantity=1
            
            # get cost for smallest quantity
            q_min=self.table.keys()[0]
            c_max=self.table[q_min]

            for q in self.table.keys():
                c=self.table[q]
                if q<q_min:
                    q_min=q
                    c_max=c
            
            for q in self.table.keys():
                c=self.table[q]
                if (quantity>=q)&(c<c_max):
                    c_max=c
                
            return c_max

    def edit_gui(self,costsdatawidget,descriptor,value):
        """
        Creating widget to edit cost-table.
        """
        # set descriptor d for edit widget to write mode
        #print 'edit_gui',self.name
        d=descriptor.copy()
        d['mode']='write'
        self.gui=gui.CostEdit(d,self.table,self.currency,costsdatawidget,
                              applybutton_command=self.modify_gui)

            
    def modify_gui(self,descriptor,coststable,costsdatawidget):
        """
        Make changes to cost table using edit_gui 
        """
        #print 'modify_costs:'
        #print ' ',descriptor['key'],'=',coststable
        self.set_value(coststable)
        descriptors,dataset=self.obj.format_costs()
        costsdatawidget.update_contents(dataset)
        #print '  self.table=',self.table

    def set_value(self,value):
        """
        Only function that can modify costs after initialization.
        """
        if self.table:
            self.table=value
        else:
            self.value=value
            
    #def on_costsdata_entry(self,costsdatawidget,costsdata):
    #    #print 'on_costsdata_entry',costsdata['Quantity']
    #    descriptor,dataset=self.obj.format_costs(costsdata['Quantity'])
    #    costsdatawidget.update_contents(dataset)
        
class Task_handling:
    """
    This class provides methods that allow a simulation object
    or module to handle tasks and processes, together with the
    message system (tasks are ordered by messages).

    Currently all management simulation objects and all modules do
    inherit this class.
    
    In order to use tasks in a simulation object or module,
    one must:
    1. provide message methods
    2. inherit this class
    3. call the init_tasks method during first initialization
    4. call the set_state_tasks method to (re)initialize pointers after a save
    5. call the get_state_tasks to clean state_image from pointers, if required
    """
    def init_tasks(self):
        """
        Must be called during first initialization.
        """

        if not hasattr(self,'tasks'):
            self.task_id_count=0  # counter to generate task ids
            # counter to generate order id for autogen tasks
            self.order_id_count=0 
            self.tasks={}           # central task db with task id as key
            self.tasks_by_client={} # task db with clientid + order id as key
            
        # seperator in order id string. Separates task id and process id
        self.order_sep='.'

        # directory that maps task types to classes
        # This directory is completed in the instance specific initializations
        self.task_classes={}

    def reset_tasks(self):
        """
        Reset task specific statistics.
        Do not destroy tasks or any relevant information concerned with tasks!
        """
        self.statistics['tasks_done']=0
        
    def set_state_tasks(self,obj,state_image={}):
        """
        This method sets pointers to functions and must be called
        during the simulation opening process.
        """
        #print 'set_state_tasks for module/object',' ',self.id

        # HERE also task-related pointers are  rebuilt.
        # this is possible because all are obj/module internal.
        # no extra rebuild_pointers calls are required
        # (after all objs are loaded as is the case for other states)
        
        self.mes_ops['order_task']=self.order_task
        self.mes_ops['kill_task']=self.kill_task
        self.mes_ops['query_task']=self.query_task
        self.mes_ops['put_info']=self.put_info

        

        if state_image:
            # managements send state image
            self.task_id_count=state_image.get('task_id_count',0)
            self.order_id_count=state_image.get('order_id_count',0)
            self.tasks=state_image.get('tasks',{})
            self.tasks_by_client=state_image.get('tasks_by_client',{})
            #print '  from state_image:',self.order_id_count
            #self.=state_image['']

        # relink tasks to (possibly newly initialized) object
        for task in self.tasks.values():
            task.obj=obj
            task.res=obj.sim.rebuild_pointer(task.res)

        # rebuild tasks_by_client database
        for id in self.tasks_by_client.keys():
            #print 'try to rebuild tasks with client id',\
            #      id,type(self.tasks_by_client[id])
            if type(self.tasks_by_client[id])==types.IntType:
                #print '  rebuild tasks',self.tasks_by_client[id]
                self.tasks_by_client[id]=\
                       self.tasks[self.tasks_by_client[id]]


            
    def get_state_tasks(self,state_image):
        """
        Removes all tasks-specific pointers from state_image.
        This is to prepare s  state_image for save operation.
        """
        #print self.id,'get_state_tasks'
        
        # remove pointers to task classes
        if state_image.has_key('task_classes'):
            del state_image['task_classes']

        for task in self.tasks.values():
            if task.res:
                if type(task.res)!=types.DictType:
                    task.res=task.res.id
            del task.obj

        for id_client in self.tasks_by_client.keys():
            self.tasks_by_client[id_client]=self.tasks_by_client[id_client].id
            
        # make sure that all message operators are removed
        #if state_image.has_key('mes_ops'):
        #    del state_image['mes_ops']

        # make sure that all task data is in the state image
        state_image['task_id_count']=self.task_id_count
        state_image['order_id_count']=self.order_id_count
        state_image['tasks']=self.tasks
        state_image['tasks_by_client']=self.tasks_by_client
          
        return state_image
    
    def order_task(self,message):
        """
        Order a new task for this module.
        All  information for the order are contained in the message.
        """

        client_id=message['from']
        order_id=message['order_id']
        # create new task
        self.task_id_count=self.task_id_count+1
        newtask=self.task_classes[message['type']](self.task_id_count,self,\
                                                   client_id,
                                                   order_id,
                                                   message['args']
                                                   )
        # store newtask in db if not already done or failed
        if (newtask.status!='done')&(newtask.status!='failed'):
            self.tasks[self.task_id_count]=newtask
            self.tasks_by_client[client_id+self.order_sep+order_id]=newtask
            return self.task_id_count
        else:
            return -1
        
            
    def del_task(self,task):
        """Delete this task"""
        if self.tasks.has_key(task.id):
            del self.tasks[task.id]
            del self.tasks_by_client[
                    task.client_id+self.order_sep+task.order_id]
            #try:
            #    del self.tasks_by_client[
            #        task.client_id+self.order_sep+task.order_id]
            #except:
            #    print 'ERROR in:del_task: did not found key',\
            #          task.client_id+self.order_sep+task.order_id
            #    for key in self.tasks_by_client.keys():
            #        print '  ',key,self.tasks_by_client[key]
            return 0
        else:
            return 1

    def query_task(self,message):
        """
        To query the status and othe information of a task in process.
        """
        task=self.get_task(message)

        # call get_info method
        if task: task.get_info()
        else:
            if self.debug>0:
                print 'ERROR in query_task: order', message['order_id'],\
                      '\n at',self.id,'from',message['from'],'not existent !!'
      
    def kill_task(self,message):
        """
        Kill task that has been ordered from the sender of this
        message with the order id of contained in this message.
        """

        task=self.get_task(message)

        # call kill method
        if task: task.kill()
        else:
            if self.debug>=5:
                print 'WARNING in kill_task: order', message['order_id'],\
                      '\n at',self.id,'from',message['from'],'not existent !!'
            
                #print '***********Tasks of',self.id,len(self.tasks)
                #for task in self.tasks.values():
                #    print '===================================='
                #    print 'task',task.id,\
                #          'from',task.client_id,'/',task.order_id,\
                #          'status=',task.status
                #    print 'in tasks_by_client ?',self.tasks_by_client.has_key(
                #        task.client_id+self.order_sep+task.order_id)

    def make_order(self,task_type,args={}):
        """
        Central function to generate an order from within
        the instansce (Autogenerated task).
        """
        self.order_id_count=self.order_id_count+1
        order_id=`self.order_id_count`

        # order task and return task_id (returns -1 if something wqent wrong)
        return self.order_task({'from':self.id,
                                'type':task_type,
                                'order_id':order_id,
                                'args':args
                                })
        
        

    def put_info(self,message):
        """
        This method is called when a task, which has been ordered by a
        process from this instance, returns information.
        Here the process which issued this order is called
        with the put info method. The process id and it's task id are
        encoded in the order_id, which must be a key in the message.
        """
        task,proc=self.get_task_proc(message['order_id'])

        # info for a process
        if proc:   proc.put_info(message['info'],message['from']) 
        else:
            if self.debug>=3:
                print 'WARNING in put_info: process', message['order_id'],\
                      '\n at',self.id,'from',message['from'],'not existent !!'

                print '***********Tasks of',self.id,len(self.tasks)
                for task in self.tasks.values():
                    print '===================================='
                    print 'task',task.id,\
                          'from',task.client_id,'/',task.order_id,\
                          'status=',task.status

    def get_task(self,message):
        """
        Get the task pointer, that has been ordered by the sender
        of the message with the order id of the message
        """
        task=None
        key=message['from']+self.order_sep+message['order_id']
        if self.tasks_by_client.has_key(key):
            task=self.tasks_by_client[key]

        return task
    
        
    def get_task_proc(self,order_id):
        """
        Read order id and return task and process which issued this order.
        """
        order_elements=string.split(order_id,self.order_sep)
        task=None
        proc=None
        if len(order_elements)>0:
            task_id=string.atoi(order_elements[0])
            if self.tasks.has_key(task_id):
                task=self.tasks[task_id]
            else:
                if self.debug>=3:
                    print 'WARNING in get_task_proc: no task:',task_id
                
        if (len(order_elements)>1)& (not(task==None )):
            proc_id=string.atoi(order_elements[1])
            if task.procs.has_key(proc_id):
                proc=task.procs[proc_id]
            else:
                if self.debug>=3:
                    print 'ERROR in get_task_proc: no process:',proc_id
            
        return task,proc

    def make_taskspage(self,parent,widgets):
        """
        Default task gui widget creation.
        Creates widgets in parent frame. All widgets are put in the
        widgets directory with a string as keyword. These widgets
        can be configured in the corresponding update method.

        May be overwritten.
        """
        header,table=self.format_tasks()
        widgets['taskstable']=gui.PanedScrolledTable(parent,header,table,
                                                     #hullwidth=450,
                                                     #hullheight=150
                                                     )
        widgets['taskstable'].pack(expand=1,fill='both',side='left')

    def update_taskspage(self,panel,widgets):
        """
        Update of widgets on task page.

        """
        header,table=self.format_tasks()
        widgets['taskstable'].make_table(header,table)


    def format_tasks(self):
        """
        Return formatted table with task information.
        """
        #print 'update_taskstable for ',self.target.id
        header=[{'key':'ID','type':'integer','width':2},
                {'key':'Status','type':'text','width':2},
                {'key':'Procs','type':'integer','width':1},
                {'key':'From ID','type':'text','width':5},
                {'key':'Order ID','type':'text','width':2}
                ]

        table=[]
        for task in self.tasks.values():
            table.append({'ID':task.id,
                          'Status':task.status,
                          'Procs':len(task.procs),
                          'From ID':task.client_id,
                          'Order ID':task.order_id,})

        return  header,table
    

    
class Task:
    """
    Class, containing common methods of a task.
    Tasks start, control and coordinate one or several processes. 
    
    """
    def __init__(self, task_id, obj, client_id,order_id, *args):
        """
        Initialization of common data structures.
        """
        
        #    print obj.id,': Init Task',task_id,\
        #          'from',client_id,'order',order_id

        # store time when task started
        self.time_start=obj.sim.time
        
        # store own id
        self.id=task_id
        
        # back link to management object
        self.obj=obj

        # store id of object that ordered this task
        self.client_id=client_id

        # store the client's internal order id
        self.order_id=order_id

        # predefine a None resource id
        self.res=None
        
        # link table to processes
        self.procs={}
        self.proc_id_count=0

        # initialize info directory, which gets passed to the
        # ordering management when status changes or during explicit
        # return info calls.
        self.info={}

        # accounting
        self.accounting_flag=1
        
        # must be set to prevent task from being killed during an algorithm
        self.operating=0 

        # make status
        self.status='waiting'
                
        # task specific initialization
        self.init_task(*args)

        
        #print 'Task of',self.obj.type,': status=',self.status

    
    def init_task(self,*args):
        """
        Task specific initializations and start action.
        The optional argument contains arguments in form of a dictionary.
        To be overwritten.
        """
        self.args=args

        # initialize and start processes:
        self.add_proc(Proc,*args)
        self.set_status('processing')

    def make_order(self,server_id,task_type,args={}):
        """
        Central function to generate an order id and issue an order
        from a task.
        """
        order_id=`self.id`
        self.obj.send_mes({'to':server_id,
                           'op':'order_task',
                           'type':task_type,
                           'order_id':order_id,
                           'args':args
                           })
        
    def put_info(self,info,sender_id=''):
        """
        Another task, which has been ordered by this task
        returns information in form of a dictionary.
        The info is read and the status of this task is changed accordingly.

        May be overwritten by specific management object.
        """
        # call task specific info treatment
        self.process_info(info)
        
        # the status of the task is set to the status of
        # the task which had been ordered.
        if self.info.has_key('status'):
            self.set_status(info['status'])
        #else:
        #    print 'WARNING:',self.obj.id,\
        #          '.proc put_info from',sender_id
        #    print '  no status in info',info
            

    def process_info(self,info):
        """
        Process newly received info from process.
        May be overwritten by specific process.
        """
        pass
    
    def get_info(self):
        """
        Send a message to the client with information about this task.
        """
        # Here just the current status is send
        self.return_info()
        
    def return_info(self,info={}):
        """Central function to send  information back to client.
        The information in info are returned, updated with the self.info
        dictionary."""
        
        returninfo={}
        returninfo.update(info)
        returninfo.update(self.info.copy())
        
        # do not return to auto-generated tasks
        if self.client_id!=self.obj.id:
            #if self.obj.id=='test.carman.experim':
            #    print 'task',self.id,'return_info status',self.status
            # make sure that info contains status
            returninfo['status']=self.status
            self.obj.send_mes({'to':self.client_id,
                               'op':'put_info',
                               'order_id':self.order_id,
                               'info':returninfo
                               })
            self.info={}
            
    def kill(self):
        """
        Start to kill all runnung processes. Task will remain untill all
        processes have terminated successfully.
        May be overwritten by specific management.
        """
        # avoid double killing
        #if (self.status!='kill')&(self.status!='killed'):
        if self.procs:
            self.set_status('kill')
            for proc in self.procs.values():
                # avoid double killing
                #if (proc.status!='kill')&(proc.status!='killed'):
                proc.kill()
        else:
            self.set_status('killed')
        

    def cleanup(self):
        """
        This function cleans up all memory traces of this task in
        the respective management.
        It is called after the task has been done, failed or killed.

        To be overwritten by specific task, if necessary.
        """
        pass
    
    def set_status(self,newstatus):
        """
        Sets status of this process, detects changes and updates task.
        """
        #if self.obj.id=='test.carman.experim':
        #        print 'set_status task',self.id,\
        #              'from',self.status,'to',\
        #              newstatus
                
        #if newstatus=='killed':
        #    print self.obj.id,': task',self.id,\
        #          'from',self.client_id,'/',self.order_id,\
        #          'Status=',newstatus

       
        if self.status!=newstatus: # status changed
            #if self.obj.id=='test.carman.experim':
            #    print 'set_status:task',self.id,newstatus
            self.status=newstatus
            if (newstatus=='done')|(newstatus=='killed')|(newstatus=='failed'):
               
                # send return info (at least status) to ordering management

                # add costs ... implement later
                #self.info['costs']=self.get_costs()

                # do some statistics
                if newstatus=='done':
                    self.obj.statistics['tasks_done']+=1
                    
                self.return_info() 

                # cleanup dbs
                self.cleanup()
                
                # remove task from management's task database
                self.obj.del_task(self) 
                #print '  ++++task',self.id,' removed!!!!!'
                return 0
        
        if self.info: # is there information to be to returned ?
            self.return_info() 

        return 0

    #def get_costs(self):
    #    """
    #    Returns costs and curreny associated with this task.
    #    This function is usually called when the task is completed
    #    and returned in the info directory to the client.
    #    """
    #    # costs do usually depend on various
    #    # task parameters plus the costs of the processes.
    #    return (self.costs_accumulated[], self.costs_accumulated.currency)

    def ignore_accounting(self):
        """
        If this function is called, the accounting for this task
        will be ignored.
        """
        self.accounting_flag=0
        for proc in self.procs.values():
            proc.ignore_accounting()
        
    def do_accounting(self):
        """
        This method does the complete accounting and other statistics
        for this task. It is assumed that the obj is a management with the
        aprobriated databases in place.

        This method is NOT automatically called when the task is completed.
        """
        # here will be space for general accounting processes of a management
        # Currently only specific accounting
        #print 'Task',self.obj.id,self.id,'self.accounting_flag',\
        #      self.accounting_flag
        if self.accounting_flag: 
            self.do_accounting_specific()
        
    def do_accounting_specific(self):
        """
        Do task specific accounting.
        To be overwritten
        """
        pass
        
        
    def add_proc(self,Newproc,*args):
        """
        Make an instance of class Proc and add it to the process table.
        This method returns the instanciation of the process.
        """
        self.proc_id_count=self.proc_id_count+1
        newproc=Newproc(self.proc_id_count,self,*args)
        self.procs[self.proc_id_count]=newproc

        return newproc

    def update_procs(self,proc=None):
        """
        This method is called from process proc if its status changed.
        The method has to decide whether the task status needs to be changed,
        if new processes need to be started. If proc has
        terminated, it needs to be deleted from the process table.

        This method may be overwritten by a specific task.
        """
        # update state of task
        self.update(proc)

        # do status changes and deal with process list.
        if proc:
            # here, terminate process how ever it finished
            if (proc.status=='done')|(proc.status=='killed'):
                # do accounting
                if proc.info.has_key('costs'):
                    self.costs_accumulated.add(info['costs'])
                self.del_proc(proc)

            elif (proc.status=='failed'):
                self.set_status('error') # make entire task unsuccessful
                self.del_proc(proc)

        
        
    def update(self,proc=None):
        """
        Update state of task when process proc has changed.

        To be overwritten.
        """
        pass
     
    def del_proc(self,proc):
        """Delete process proc from task's process table"""

        #print self.obj.id,': No of Processe of task',self.id,':',\
        #      len(self.procs)

        if self.procs.has_key(proc.id):
            #print 'del_proc',self.id,'.',proc.id,'on',self.obj.id,\
            #      len(self.procs),not self.operating,self.status
            
            del self.procs[proc.id]
                
            if not self.procs:            # no more processe ?
                # if task is not operating, conclude task automatically.
                if not self.operating:
                    # set definitive status according to the present status
                    # => task terminated...and will be destroyed
                    if   self.status=='processing': self.set_status('done')
                    elif self.status=='waiting'   : self.set_status('done')
                    elif self.status=='error'     : self.set_status('failed')
                    elif self.status=='kill'      : self.set_status('killed')
                else:
                    print 'WARNING: del_proc: tried to kill',proc.id,\
                          'while operating in',self.status
            

class Proc:
    """
    Class, containing common methods of a process.
    
    """
    def __init__(self,id,task,*args):
        """
        Initialization of common data structures
        """

        #print task.obj.id,': Init Proc',task.id,'.',id
        
        self.id=id     # unique process id
        self.task=task # pointer to task object
        self.status='waiting'
        

        self.res_id=''   # id of connected resource
        self.order_id='' # id of task order for this resource
        self.info={} # info table to be returned to task.

        # accounting
        self.accounting_flag=1
        
        # process specific initialization
        self.init_proc(*args)

    def init_proc(self,*args):
        """
        Process specific initializations and start action.
        The optional argument contains arguments from the main programme.
        To be overwritten.
        """
        self.args=args

        # do something...
        self.set_status('processing')
        #print 'Proc:',self.id,'Status:',self.status
        self.set_status('done')
        #print 'Proc:',self.id,'Status:',self.status


    def make_order(self,res_id,task_type,args={}):
        """
        Central function to generate an order id and issue an order
        for processes.
        """

        #print 'proc:make_order for ',server_id

        # compose order_id of task id and process id
        self.order_id=`self.task.id`+self.task.obj.order_sep+`self.id`
        self.res_id=res_id
        self.task.obj.send_mes({'to':res_id,
                                'op':'order_task',
                                'type':task_type,
                                'order_id':self.order_id,
                                'args':args
                                })
        
    def put_info(self,info,sender_id=''):
        """
        A task, which has been ordered by this process 
        returns information in form of a dictionary.
        The info is read and the status of the process is changed accordingly.
        """

        # transfer received info to task info.
        # this may overwrite information!
        
        
        self.info.update(info)
        if info.has_key('status'):
            del info['status']
        #print self.task.obj.id,'put_info in proc',info,self.info
        
        # Here the status of the process is simply set to the status of
        # the task which had been ordered.
        self.process_info(info)
        if self.info.has_key('status'):
            self.set_status(self.info['status'])
        #else:
        #    print 'WARNING:',self.task.obj.id,\
        #          '.proc put_info from',sender_id
        #    print '  no status in info',info

    def process_info(self,info):
        """
        Process newly received info from client.
        May be overwritten by specific process.
        """
        pass
    
    def kill(self):
        """
        Terminate this process.
        May be overwritten by specific process.
        """
        if self.res_id:
            # avoid double killing
            if (self.status!='kill')&(self.status!='killed'):
                # send task of resource the kill command
                self.task.obj.send_mes({'to':self.res_id,
                                        'op':'kill_task',
                                        'order_id':self.order_id,
                                        })
                
                self.set_status('kill')
        else:    
            self.set_status('killed')
        
    def set_status(self,newstatus):
        """
        Sets status of this process, detects changes and updates task.
        """
        #if self.debug>=5:
        # print self.task.obj.id,': Status of proc',self.task.id,'.',self.id\
        #          ,'=',newstatus

        # update task only if status changed and process is not killed 
        # make sure that a process in kill status can only
        # turn into a killed status.

        #if (self.status=='kill')&(newstatus=='killed'):
        #    print 'detected kill->killed'
            
        if (self.status!=newstatus): #&(self.status!='kill'):
            #if self.task.obj.id == 'test.carman.experim':
            #    print 'set_status:proc',self.task.id,'.',self.id,newstatus
                
            self.status=newstatus

            # implement later...
            # check if there where costs involved
            #if info.has_key('costs'):
            #    self.costs_accumulated.add(info['costs'])
            #    del info['costs']
            #    self.info['costs']=self.get_costs()
                
            self.task.update_procs(self)
            
                
        
        #elif (self.status=='kill')&(newstatus=='killed'):
        #    self.status=newstatus
        #    self.task.update_procs(self)
        return 0
        #if self.status==newstatus:
        #    return 0
        #else:
        #    print 'ERROR in proc.set_status in',\
        #          self.task.obj.id,'task',self.task.id,'proc',self.id
        #    print '  could not set status',self.status,'to',newstatus
        #    return 1

    def ignore_accounting(self):
        """
        If this function is called, the accounting for this task
        will be ignored.
        """
        self.accounting_flag=0
        
    def do_accounting(self):
        """
        This method does the complete accounting and other statistics
        for this task. It is assumed that the obj is a management with the
        aprobriated databases in place.

        This method is  automatically called when the process is completed
        in ANY way (killed, done,...)
        """
        # here will be space for general accounting processes of a management
        # Currently only specific accounting
        #print 'Task',self.obj.id,self.id,'self.accounting_flag',\
        #      self.accounting_flag
        if self.accounting_flag: 
            self.do_accounting_specific()

    def do_accounting_specific(self):
        """
        Do process specific accounting.
        To be overwritten
        """
        pass
    
    #def get_costs(self):
    #    """
    #    Returns costs and curreny associated with this process.
    #    This function is usually called when the task is completed
    #    and returned in the info directory to the client.
    #    """
    #    # costs do usually depend on various
    #    # task parameters plus the costs of the processes.
    #    return (self.costs_accumulated[], self.costs_accumulated.currency)
    


class Module(Task_handling,Costs_handling):
    '''
    Class for common functions and states of modules
    '''
    def init_common(self,obj):
        '''
        Initialization, common to all modules.
        Function called by specific module only ONCE during first
        initialization.
        '''
        # back-link with simulation track object
        self.obj=obj
        
        # copy simulation environment in module object
        self.sim=self.obj.sim

        # init debug level
        self.debug=self.obj.sim.debug
        
        # this flag will be set if the module is to be deleted
        # as soon as possible.
        self.del_flag=0
        
        # message FIFO buffer
        self.messages=[]
        self.read_mes_on_receive=1 # reads and evaluates message immediately 

        # default management simulation object
        self.man_id=self.obj.man_id

        # default parameters
        self.init_params()
        self.set_params()

        # make sure that management sim obj is in simulation
        self.obj.sim.append_obj(self.man_id,link_flag=0)
        
        

        # statistics data initialization reset will be called by object
        #self.init_statistics()
        self.statistics={}
        self.reset_statistics()


        # init variables for messaging
        self.mes_ops={'error_mes':self.error_mes}
        
        # tasks
        self.init_tasks()
        self.set_state_tasks(obj)
        # costs stuff, here only variable initialization
        # very dependent on module, needs to be updated 

        
        # Costs stuff
        self.costs={}
        #self.cost_functions={}
        self.init_costs()

        
        
        # init states that need to be set also after reload
        #self.set_state(self.obj)
        # done in __init__ after ALL specific initializations
        
        
    def init_specific(self):
        '''
        Still necessary , module specific init at the beginning ??
        Here goes module specific initializations; to be overwritten'''
        pass

    def get_version(self):
        return self.obj.get_version()
    
    def set_state(self,obj):
        '''
        
        This function is called during the opening process of a simulation
        file. Here goes all the initializations that are COMMON to all
        modules and that gets lost during the save process (essentially
        pointers to functions).  It is also called during the module
        initialization. Hence, it is also possible to define atributes
        that has been added in the new version, but that need to be
        initialized for objects that have been saved under an older
        version.

        '''
        #print 'set_state: of',self.id
        
        # pointer to administrating simulation object
        self.obj=obj
        self.debug=self.obj.sim.debug
        
        # point to sim environment
        self.sim=self.obj.sim
        
        
        # management object id default
        # self.man_id=self.obj.man_id   

        # Message opeartors, to be extended by
        # module-specific call set_state_mod
        self.mes_ops={'error_mes':self.error_mes}

        # tasks
        self.init_tasks()
        
        # task settings
        self.set_state_tasks(self)

                
        # call module specific settings
        self.set_state_mod()

        
        
    def set_state_mod(self):
        '''
        This function is called during the opening process of a simulation
        file. Here goes all the stuff that gets lost during the save
        process (essentially pointers to functions).  It is also called
        during the module initialization. Hence, it is also possible to
        define atributes that has been added in the new version, but that
        need to be initialized for objects that have been saved under an
        older version.

        Usually overwritten by the specific module.
        '''
        pass

    
    def get_state(self,state_image):
        '''
        This function removes all pointers to function in the state_image
        dictionary and returns a  state_image that is ready for pickling.
        The pointers removed by this function must be restored in the
        set_state function while opening the simulation file.
        '''
        #print self.id,'get_state'

        # removes all module-specific pointers from state_image
        state_image=self.get_state_mod(state_image)

        # removes all pointers to functions concerned with tasks
        state_image=self.get_state_tasks(state_image)
        
        if state_image.has_key('sim'):
            del state_image['sim']

        if state_image.has_key('obj'):
            del state_image['obj']
        
        if state_image.has_key('mes_ops'):
            del state_image['mes_ops']
            
        

        # remove cost functions
        #self.cost_functions={}
        
        
        
        #for key in state_image.keys(): print '  state:',key
        #for key in self.__dict__.keys(): print '  dict:',key
        return state_image

    def get_state_mod(self,state_image):
        '''
        This function removes all module-specific pointers from the
        state-image, see get_state function.
        To be overwritten by module-specific function.
        '''
        return state_image


            
    def init_params(self,param={}):
        """
        Initialize parameters. Is called only once during init_common.
        Here all constant attributes should be defined.
        To oververwrite.
        """
        self._params={}
        
    def set_params(self,param={}):
        """
        Set arbitrary parameters and make them effective, if possible.
        To oververwrite.
        """
        pass

    def write_if_status_edit(self):
        """
        Returns 'write' only when the status is 'edit'
        """
        if self.status=='edit':
            return 'write'
        else:
            return 'read'
        
    def init_costs(self):
        """
        Define default costs here.

        To be overwritten.
        """
        self.set_costs({'estim_init':{'name':'Estimated initial costs',
                                      'table':{1:     80.0*10**3,
                                               10:    40.0*10**3,
                                               100:   20.0*10**3,
                                               1000:  15.0*10**3,
                                               10000: 10.0*10**3,
                                               100000: 5.0*10**3,},
                                      'mode':'write'},
                        'estim_pa':{'name':'Estimated costs per year',
                                    'table':{1:      1000.0,
                                             10:     900.0,
                                             100:    800.0,
                                             1000:   700.0,
                                             10000:  500.0,
                                             100000: 100.0,},
                                    'mode':'write'},
                        },
                       lifetime_estim=10)
        
    def do_costs_current(self):
        """
        Calculates current costs and modifies the costs dictionary.
                
        """
        self.do_costs_current_mod()
        
    def do_costs_current_mod(self):
        """
        Calculates module spacific current costs
        and modifies the costs dictionary.
                
        """
        pass

    def make_costspage_specific(self,parent,widgets):
        """
        Costs page gui creation, specific to all modules.
        """
        # ' Create a apply-buttonbox'
        #type=self.obj.sim.split_id(self.id)['type']

        widgets['buttonbox'] = gui.ButtonArray(parent,
                     button_array=\
                     [('Apply costs to all modules of this type',
                       self.on_apply_costs_to_all,\
                       {'costsdatawidget':widgets['costsdata']})],
                       defaultbutton='Apply costs to all modules of this type',
                                               )

        # limit size of costdata widget
        #widgets['costsdata'].set_size(height=240)

        widgets['buttonbox'].pack(expand=1,side = 'bottom', fill = 'x')

        
    def format_costs_specific(self,quantity,descriptors,table):
        """
        Cost formatting, specific to all modules.
        """
        descriptors.append({'key':'Quantity',
                            'type':'integer',
                            'width':1,
                            'valid':{'max':10**8},
                            'mode':'read',
                            })
        table['Quantity']=quantity
        # do module specific stuff
        self.format_costs_mod(quantity,descriptors,table)

    def format_costs_mod(self,quantity,descriptors,table):
        """
        Module specific cost formatting.
        """
        pass

    def format_params(self):
        """
        Returns describtor and dataset of parameters.
        """
        return [],{}
    
    def make_panel(self,panel,**kw):
        """
        Initialize widgets on panel.
        """
        # create standard layout with pages
        panel.make_standardpanel()

        # here one can add pages like
        #panel.make_page('Info',self.make_infopage,self.update_infopage)

    def make_performancepage(self,parent,widgets):
        """
        Default performance gui widget creation.
        Creates widgets in parent frame. All widgets are put in the
        widgets directory with a string as keyword. These widgets
        can be configured in the corresponding update method.

        May be overwritten.
        """
        
        descriptor,dataset=self.format_statistics()
        widgets['statisticsdata']=gui.DatasetEditor(descriptor,dataset,parent)
        widgets['statisticsdata'].pack(expand=1,fill='both')

    def update_performancepage(self,panel,widgets):
        """
        Update of widgets on performance page.
        """
        descriptor,dataset=self.format_statistics()
        widgets['statisticsdata'].update_contents(dataset)

    
        
    def reset_statistics(self):
        """
        Reset or initialize statistical variables.
        """
        # Task stuff
        self.reset_tasks()
        self.reset_statistics_mod()

    def reset_statistics_mod(self):
        """
        Reset or initialize module specific statistical variables.
        """
        pass

    def format_statistics(self):
        """
        Return descriptor and dataset contents for statistical data
        about the target object.

        To be overwritten.
        """
        descriptor=[{'key':'Efficiency',
                     'type':'text',
                     'width':1,
                     'mode':'read', # or 'write' if editable
                     'units':{'%':1}},
                    ]
        dataset={'Efficiency':100.0}
        self.format_statistics_specific(descriptor,dataset)    
        return descriptor,dataset

    def format_statistics_specific(self,descriptor,dataset):
        """
        Return descriptor and dataset contents for statistical data
        about the specific module.

        To be overwritten.
        """
        pass
        
    def show_info(self,debuglevel=0):
        print 'Module',self.id

    def save_info(self,sep=' '):
        file_name =tkFileDialog.SaveAs(
                initialdir=self.sim.simdir,
                filetypes=[('Text files','*.txt'),('All files','*')]).show()

        try:
            file=open(file_name,'w')
            for row in self.format_info(sep): file.write(row+'\n')
            file.close()
            return 0
        except:
            #print 'WARNING in save: could not open',file_name
            return 1

    def print_info(self,sep=' '):
        print '----------------------------------------------------------'
        print 'Print info at simulation time [hh:mm:ss]',\
              self.sim.seconds_to_string(self.sim.time),'\n'
        
        for row in self.format_info(sep):print row
        self.show_info()

    def format_info(self,sep):
        """
        Returns a string list with formatted info.
        """
        il=[]
        dataset_func=gui.DataSet()
        il.append('Module'+sep+self.id+'\n')

        il.append('\n Performance:\n')
        descriptors,dataset=self.format_statistics()
        il+=dataset_func.format_dataset(descriptors,dataset)

        il.append('\n Costs:\n')
        descriptors,dataset=self.format_costs()
        il+=dataset_func.format_dataset(descriptors,dataset)

        descriptors,dataset=self.format_params()
        if descriptors:
            il.append('\n Parameters:\n')
            il+=dataset_func.format_dataset(descriptors,dataset)

        self.format_info_specific(il)
        #il.append()
        return il

    def format_info_specific(self,il,sep=' '):
        """
        Returns a string list with formatted info.
        """
        pass
    
    def __getstate__(self):
        '''
        Function, called by the picle process. This function must
        return the pickel-ready state of the module
        '''
        return self.get_state(self.__dict__)

    #def __setstate__(self,states):
    #    '''
    #    Function, called by the picle process. This function must
    #    return the pickel-ready state of the module
    #    '''
    #    print '__setstate__ for module',states['id'],'\n  ',self
    #    for state in states.keys(): print '  ',state
        
    def put_mes(self,message):
        '''Function called for putting messages in this module'''
        self.messages.append(message)
        if self.debug>=5:
            print 'put message to module:',self.id
            print 'message:\n',message
            
        if self.read_mes_on_receive:
            self.read_mes()
            
    def send_mes(self,message):
        '''Function used to send messages to objects or directly to modules'''
        message['from']=self.id
        id_elements=self.obj.sim.split_id(message['to'])

        #if message['to']=='base.userman.experim':
        #    print self.id,',send_mes to id_elements=',id_elements
        #    print 'message=',message
        
        if id_elements['mod']!=None:
            # send message directly to module
            id_elements['mod'].put_mes(message)
        elif id_elements['obj']!=None:
            # send mesage to simulation object
            id_elements['obj'].put_mes(message)
        else: # destination is not a valid simulation object or module
            if self.debug>=3:
                print 'ERROR: send_mes:',\
                      'to-address',message['to'],\
                      'is not a valid simulation object or module!'
            return 1
        return 0

        
    def read_mes(self):
        '''Function called to read and evaluate new messages'''

        while self.messages:
            message=self.messages.pop(0)
            #print '--------------------------------------------------'
            #print 'process message:\n',message
            # check if message contains an operation, defined in the
            # public operators map.
            #if hasattr(self,'mes_ops'):
            #try: # this is  test operation !!
            #    a=self.mes_ops.has_key(message['op'])
            #except:
            #    print 'No mes_ops !!!:',self.id
                
            if self.mes_ops.has_key(message['op']):
                # execute operation
                self.mes_ops[message['op']](message)
            else:
                # send back an error message
                self.send_mes({'to':message['from'],
                               'from':self.id,
                               'op':'error_mes',
                               'error':'unknown operator '+message['op'],
                               'old_message':message}
                              )
            #print '--------------------------DONE--------------------'
                        
        
            
    def error_mes(self,message):
        print 'ERROR occured in module:',self.id
        print 'Message to',message['from'],\
              'unsuccessfully delivered.\n error: ',message['error']

    
    
class Obj:
    '''
    Class for common functions and states of simulation objects
    '''
    def __init__(self, type, state_image, sim):
        # default version is the current version of its
        # when an object is loaded from an older version, the version
        # will be replaced by an older version,
        # The version should be upgraded 
        # in the set_state or get_state functions and subfunctions.
        self.__version__=sim.__version__
        
        self.name=type # should be replaced by a more human readable string
        self.info='No info text available' # short info text
        
        # default instructions for simulator, to overwrite in specific object
        self.iterate_interval=0.0 # in seconds simulation time 
        self.output_interval=0.0  # in seconds simulation time 
        self.gui_flag=0    # =1: append and use ONLY in gui simulation mode

        # for simulation objects type and id are currently identical
        self.type=type
        self.id=type

        # An object has no mother object, unlike modules
        # this is used for generalization reasons
        # to indicate that it is not a module
        self.obj=None 
        
        self.sim=sim
        self.debug=sim.debug

        # Message stuff
        self.messages=[] # message buffer
        self.read_mes_on_receive=1 # reads and evaluates message immediately 
        
        # Operators, that are allowed to be executed directly
        # by message contents. This table of operators, mapping key-strings
        # to functions, are expanded in specific initializations.

        # Message opeartors for simulation object
        self.mes_ops={'error_mes':self.error_mes} 

        
        # counters administered by mainloop of simenv
        self.iterate_count=0.0
        self.update_count=0.0
        
        # indicates that rescalling occurred (for gui use only)
        #self.netmon=None
        self.rescaled_flag=0
        
        # call specific initialization of simulation object
        self.init_obj()

        # do further initializations only if the object belongs to
        # the correct simulation mode
        #print '__init__',self.id
        #print '  gui_flag=',self.gui_flag
        #print '  sim.mode=',sim.mode
        #if (not self.gui_flag)|(self.gui_flag & (sim.mode=='gui')):
        if 1: # all ojects will be completely initialized   
            # check if obj  can contain multiple modules
            #if hasattr(self,'add_mod'):
            if self.__dict__.has_key('Mod'):
                self.multimod_flag=1
                self.mods={}
                self.mod_no_current=0
            else:
                self.multimod_flag=0

            # initialize statistics
            self.statistics={}
            self.reset_statistics()
        
            # call state reconstruction, if state image is available
            if state_image!=None:
                #print self.id,'start setting state_image'
                # is there more than just the object type
                if len(state_image)>1: 
                    self.set_state_image(state_image)

    def get_version(self):
        return self.__version__
    
    def get_state_image(self):
        '''Get the state image of this simulation object for saving'''
        # print self.id,'get_state_image'
        # get specific states from simulation object
        state_image=self.get_state()
        
        # The following assignments complete the state image  
        # with states that are common to all  simulation objects.
        # They do not need to be specified in self.get_state of each object!!

        state_image['version']=self.__version__
        state_image['type']=self.type
        state_image['messages']=self.messages
        state_image['statistics']=self.statistics
        #for message in self.messages: print '  message',message['from']
        #if not state_image.has_key('type'):
        #    state_image['type']=self.type
        #if not state_image.has_key('version'):
        #    state_image['version']=self.__version__
        #if state_image.has_key('mes_ops'):
        #    del state_image['mes_ops']
            
        return state_image
        
    def get_state(self):
        '''Get the object specific state state image  for saving'''
        # returns state image of the current state as a tuple.
        # To be overwritten
        si={}
        self.get_state_specific(si)
        return si

    def get_state_specific(self,state_image):
        """
        For specific object. To be overwritten.
        This function is called with state_image as argument
        which will be completed with specific data.
        """
        pass
    
    def set_state_image(self,state_image):
        '''Sets the state image of this simulation object while opening'''
        #print self.id,'set_state_image'
        # set common states
        
        # it is left to object specific set_state to upgrade the version
        self.__version__=state_image['version']
        
        self.type=state_image['type']
        self.messages=state_image['messages']
        #for message in self.messages: print '  message',message['from']
        # set object specific states
        self.statistics=state_image['statistics']
        
        self.set_state(state_image)
        self.debug=self.sim.debug
        
    def close(self):
        """
        Does necessary action before this simulation object is
        deleted from memory.

        It can be assumed that the state of the object has been
        saved. So what remains is to clear up other things, essentially
        GUI stuff.
        """
        pass
    
    def set_state(self,state_image):
        # overwrite with an algorithm to restore the object's state
        # from the state_image data object
        self.set_state_specific(state_image)
        
    def set_state_specific(self,state_image):
        """
        For specific object. To be overwritten
        """
        pass

    
    def init_obj(self):
        # Method is called for the final initialization
        # state image, simenv and maybe some states are already initialized
        # Overwrite this method
        pass

    def get_mod_id(self,mod):
        return self.sim.get_id(self.type,mod.no)
    
    def try_link_with(self,type):
        if self.sim.objs.has_key(type):
            obj=self.sim.objs[type]
            # link only to object that are actively used in current
            # simulation mode.
            # at the moment this means not to link with a gui object
            # if simulation is not in gui mode.
            if (not obj.gui_flag)|( obj.gui_flag& (self.sim.mode=='gui')):
                return obj
            else:
                return None
        else:
            return None

    
    
    def link(self):
        """
        Make links to other simulation objects.
        Use: try_link_with method.
        """
        pass

    def rebuild_pointers(self):
        """
        Rebuild pointers that have been removed before saving.
        This function is called AFTER all objects have been
        initialized and cross-linked.
        """
        pass
    
    def change_time_sample(self,time_sample):
        """
        Sampling time has changed. Do parameter changes if necessary.
        """
        pass
    def put_mes(self,message):
        '''Function called for putting messages in this object'''
        self.messages.append(message)
        if self.debug==5:
            print 'put message to simulation object:',self.type
            print 'message:\n',message
        if self.read_mes_on_receive==1:
            self.read_mes()

    def send_mes(self,message):
        '''Function used to send messages to objects or directly to modules'''

        message['from']=self.type
        id_elements=self.sim.split_id(message['to'])
        if id_elements['mod']!=None:
            # send message directly to module
            id_elements['mod'].put_mes(message)
        elif id_elements['obj']!=None:
            # send mesage to simulation object
            id_elements['obj'].put_mes(message)
        else: # destination is not a valid simulation object or module
            if self.debug>=3:
                print 'ERROR: send_mes:',\
                      'to-address',message['to'],\
                      'is not a valid simulation object or module!'
            return 1
        return 0
    
    def read_mes(self):
        '''Function called to read and evaluate new messages'''
        
        while self.messages:
            message=self.messages.pop(0)
            #if self.id=='base.userman.experim':
            #    #print '--------------------------------------------------'
            #    print 'process message:\n',message

            # check if message contains an operation, defined in the
            # public operators map.
            if self.mes_ops.has_key(message['op']):
                # execute operation
                self.mes_ops[message['op']](message)
                
            else:
                # send back an error message
                self.send_mes({'to':message['from'],
                               'op':'error_mes',
                               'error':'unknown operator',
                               'old_message':message}
                              )
            #print '--------------------------DONE--------------------'
                
        self.messages=[]
        
    def error_mes(self,message):
        print 'ERROR occured in simulation object:',self.type
        print 'Message to',message['from'],\
              ' unsuccessfully delivered.\n error: ',message['error']

    #def get_report(self,tab=' '):
    #    """
    #    Returns string with report.
    #    """
    #    return ''
    
    def rescaled(self,canvas):
        """
        The canvas or netmon has been rescaled. Do something, if necessary.
        """
        self.rescaled_flag=1
    
    def get_panel(self,panelframe):
        """
        Initialize and return panel widget.
        """
        return gui.Panel(self,panelframe)

    def reset_statistics(self):
        """
        Reset or initialize statistical variables.

        To be overwritten.
        """
        
        if self.multimod_flag:
            for mod in self.mods.values():
                mod.reset_statistics()
                
        self.reset_statistics_specific()

    def reset_statistics_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular carrier type.

        To be overwritten.
        """
        pass
    
    def report_statistics(self):
        """
        Return descriptor and dataset contents for statistical data
        about the target object.
        """
        #print 'format_statistics',self.id
        descriptor=[]
        dataset={}
        dataset_counter={}
        dataset_sum={}
        if self.multimod_flag:
            descriptor.append({'name':'Number',
                               'key':'number',
                               'type':'integer',
                               'printformat':'%9d',
                               'width':1,
                               'valid':{'max':10**8},
                               'mode':'read'},
                              )
            dataset['number']=len(self.mods)
            
            # cheap trick: ask modules for data and return average values
            if len(self.mods)>0:
                # init a conter and sum for reach key in the dataset
                
                for mod in self.mods.values():
                    self.cumulate_mod_results(mod.format_statistics,descriptor,
                                              dataset_sum,dataset_counter)

                # average all data
                for key in dataset_sum.keys():
                    #print '  ',key,dataset_sum[key]/dataset_counter[key]
                    dataset[key]=dataset_sum[key]/dataset_counter[key]
        else:
            descriptor,dataset=self.format_statistics()
            
        self.report_statistics_specific(descriptor,dataset,
                                        dataset_sum,dataset_counter)
            
        return descriptor,dataset
    
    def cumulate_mod_results(self,function,descriptor,sum,counter):
        """
        Helper method to cumulate results from modules by appying the
        given function.
        Data, if numerical, will be added to the value in sum dictionary with
        the corresponding key. The counter with the corresponding key will
        also be increased.
        """
        #print 'format_statistics',self.id
        descriptor_mod,dataset_mod=function()
        for descript in descriptor_mod:
            key=descript['key']
            
            # can it be treated as a number
            if descript['type'] in ('real','integer','int'):
                # check if there is already a counter
                if counter.has_key(key):
                    counter[key]+=1 # increase
                    sum[key]+=dataset_mod[key]
                else:
                    descriptor.append(descript)
                    counter[key]=1 # initialize
                    sum[key]=dataset_mod[key]
                        
                

    def report_statistics_specific(self,descriptor,dataset,
                                   dataset_sum,dataset_counter):
        """
        Return object specific descriptor and dataset contents for report on
        statistical data.
        """
        pass
    
    def report_costs(self):
        """
        Return descriptor and dataset contents for  report on costdata.
        """
        descriptor=[]
        dataset={}
        dataset_counter={}
        dataset_sum={}
        if self.multimod_flag:
            # cheap trick: ask modules for data and return average values
            if len(self.mods)>0:
                # init a conter and sum for reach key in the dataset
                
                for mod in self.mods.values():
                    self.cumulate_mod_results(mod.format_costs,descriptor,
                                              dataset_sum,dataset_counter)

                #print 'report_costs',self.id
                for descript in descriptor:
                    key=descript['key']
                    if dataset_sum.has_key(key):
                        # accumulate these costs
                        if key in ('estim_init',
                                   'estim_pa',
                                   'revenue',
                                   'revenue_per_h'):
                            dataset[key]=dataset_sum[key]
                        # ignore these items
                        elif key in ('lifetime_estim','quantity','Quantity'):
                            pass
                        else: # average
                            dataset[key]=dataset_sum[key]/dataset_counter[key]
                            
                    # take sum if unit is currency, otherwise average
                    #if dataset.has_key(key):
                    #    if descript.has_key('units'):
                    #        if (self.sim.currency not in \
                    #               descript['units'].keys())&\
                    #               (self.sim.currency+'/a' not in \
                    #                descript['units'].keys()):
                    #            dataset[key]=dataset_sum[key]/\
                    #                          dataset_counter[key]
                    #        else:
                    #            dataset[key]=dataset_sum[key]
                    #    else:
                    #        dataset[key]=dataset_sum[key]/\
                    #                      dataset_counter[key]
                            
        else:
            descriptor,dataset=self.format_costs()
            
        self.report_costs_specific(descriptor,dataset,
                                   dataset_sum,dataset_counter)
        
        return descriptor,dataset

    def report_costs_specific(self,descriptor,dataset,
                              dataset_sum,dataset_counter):
        """
        Return object specific descriptor and dataset contents for report on
        costs data.
        """
        pass


    def report_params(self):
        """
        Return descriptor and dataset contents for parameters
        about the object.
        """
        #print 'format_statistics',self.id
        descriptor=[]
        dataset={}
        dataset_counter={}
        dataset_sum={}
        if self.multimod_flag:
            
            # cheap trick: ask modules for data and return average values
            if len(self.mods)>0:
                # init a conter and sum for reach key in the dataset
                
                for mod in self.mods.values():
                    self.cumulate_mod_results(mod.format_params,descriptor,
                                              dataset_sum,dataset_counter)

                # average all data
                for key in dataset_sum.keys():
                    #print '  ',key,dataset_sum[key]/dataset_counter[key]
                    dataset[key]=dataset_sum[key]/dataset_counter[key]
        else:
            descriptor,dataset=self.format_params()
            
        #self.report_params_specific(descriptor,dataset,
        #                                dataset_sum,dataset_counter)
            
        return descriptor,dataset

    
    def format_params(self):
        """
        Return descriptor and dataset contents for parameters.
        """
        descriptor=[]
        dataset={}
        return descriptor,dataset



    def save_info(self,sep=' '):
        file_name =tkFileDialog.SaveAs(
                initialdir=self.sim.simdir,
                filetypes=[('Text files','*.txt'),('All files','*')]).show()

        try:
            file=open(file_name,'w')
            for row in self.format_info(sep): file.write(row+'\n')
            file.close()
            return 0
        
        except:
            #print 'WARNING in save: could not open',file_name
            return 1
        

    def print_info(self,sep=' '):
        print '----------------------------------------------------------'
        print 'Print info at simulation time [hh:mm:ss]',\
              self.sim.seconds_to_string(self.sim.time),'\n'
        
        for row in self.format_info(sep):print row
        self.show_info()

    def format_info(self,sep):
        """
        Returns a string list with formatted info.
        """
        il=[]
        dataset_func=gui.DataSet()
        il.append('Object'+sep+self.id+'\n')

        il.append('\n Performance:\n')
        descriptors,dataset=self.format_statistics()
        il+=dataset_func.format_dataset(descriptors,dataset)

        il.append('\n Costs:\n')
        descriptors,dataset=self.format_costs()
        il+=dataset_func.format_dataset(descriptors,dataset)


        descriptors,dataset=self.format_params()
        if descriptors:
            il.append('\n Parameters:\n')
            il+=dataset_func.format_dataset(descriptors,dataset)
            
        

        self.format_info_specific(il)
        #il.append()
        return il

    def format_info_specific(self,il,sep=' '):
        """
        Returns a string list with formatted info.
        """
        pass
    
    def lock(self,switch='on'):
            if switch=='on':
                self.threadlock=thread.allocate_lock()
                # wait until maincycle is terminated
                while self.sim.maincycle_flag: pass
                self.threadlock.acquire()
                  
            elif switch=='off':
                self.threadlock.release()
               
    def show_info(self):
        """
        Print state of selected module in terminal window.
        """
        print 'Sim-Object:',self.id
    
    def __getstate__(self):
        return {'type':self.type}

    def __setstate__(self,state):
        self.type=state['type']

class Vector(UserArray):
    """
    This is a class extended 1D version of NumPy's array build-in functions.
    Major feature: it allows deleting elements.
    """
    def __init__(self, data=None, typecode=None, copy=1, savespace=0):
        if data==None:
            data=[]# initialise with empty matrix
        #if len(data)==0:
        #    data=[]# initialise with empty matrix
            
        if type(data) is types.StringType:
            data = _convert_from_string(data)
        UserArray.__init__(self,data,typecode, copy, savespace)
        #if len(self.array.shape) == 1:
        #    self.array.shape = (1,self.array.shape[0])
        self.shape = self.array.shape

    def append(self,subvec):
        """
        Appends sub vector. 
        """
        if self.is_empty():
            self._modify_array(subvec.array)
        else:
            self._modify_array(concatenate((self.array,subvec.array),0))

    def append_array(self,array):
        """
        Appends sub vector. 
        """
        if self.is_empty():
            self._modify_array(array)
        else:
            self._modify_array(concatenate((self.array,array),0))
        
    def is_empty(self):
        return len(self.array)==0
    
        
    def _modify_array(self,data,typecode=None, copy=1, savespace=0):
        delattr(self,'array')
        self.array=array(data,typecode=None,copy=copy, savespace=savespace)
        self.shape = self.array.shape
        self._typecode = self.array.typecode()
    def __delitem__(self,ind):
        """
        index of the corresponding
        vector element to be deleted.
        """
        if ind!=0:
            self._modify_array(concatenate((
                self.array[0:ind],
                self.array[ind+1:len(self)]
                    )))
        else:
            self._modify_array(self.array[ind+1:len(self)])

    def put(self,inds,subvector):
        put(self.array,inds,subvector.array)

    def take(self,inds):
        return self._rc(take(self.array,inds))

    def put_array(self,inds,subarray):
        put(self.array,inds,subarray)

    def take_array(self,inds):
        return take(self.array,inds)

    def __getstate__(self):
        return {'array':self.array}
    
    def __setstate__(self,state):
        self.array=state['array']
        self.shape = self.array.shape
        self._typecode = self.array.typecode()


    
    
class Matrix(Matrix):
    """
    This is a class extended 2D version of NumPy's array build-in functions.
    Major feature: it allows deleting rows and columns.
    """
    def __init__(self, data=None, typecode=None, copy=1, savespace=0):

        if data==None:
            data=[[]]# initialise with empty matrix
        #elif len(data)==0:
        #    data=[[]]# initialise with empty matrix
        #elif len(data)==1:
        #    if len(data[1])==0:
        #        data=[[]]
            
        if type(data) is types.StringType:
            data = _convert_from_string(data)
        UserArray.__init__(self,data,typecode, copy, savespace)
        if len(self.array.shape) == 1:
            self.array.shape = (1,self.array.shape[0])
        self.shape = self.array.shape
        
    def _modify_array(self,data):
        delattr(self,'array')
        self.array = array(data,copy=1)
        self.shape = self.array.shape
        self._typecode = self.array.typecode()
    def append_col(self,array):
        """
        Appends column. Column is a 1D array
        """
        col=transpose([array])
        if self.is_empty():
            self._modify_array(col)
        else:
            self._modify_array(concatenate((self.array,col),1))
        
    def append_row(self,array):
        """
        Appends row. Row is a 1D array
        """
        row=[array]
        if self.is_empty():
            self._modify_array(row)
        else:
            self._modify_array(concatenate((self.array,row),0))

    def is_empty(self):
        if self.array.shape==(1,0):
            if len(self.array[0])==0:
                return 1
        return 0
        
    def del_row(self,ind):
        """
        Deletes column with index ind.
        """
        if ind!=0:
            self._modify_array(concatenate((
                self.array[0:ind],
                self.array[ind+1:len(self)]
                    )))
        else:
            self._modify_array(self.array[ind+1:len(self)])


            
    def del_col(self,ind):
        """
        Deletes column with index ind.
        """
        if ind!=0:
            self._modify_array(concatenate((
                self.array[:,0:ind],
                self.array[:,ind+1:self.shape[1]]
                    ),1))
        else:
            self._modify_array(self.array[:,ind+1:self.shape[1]])

    
    def del_row(self,ind):
        """
        Deletes column with index ind.
        """
        if ind!=0:
            self._modify_array(concatenate((
                self.array[0:ind],
                self.array[ind+1:len(self)]
                    )))
        else:
            self._modify_array(self.array[ind+1:len(self)])

    def __setitem__(self, ind, value):
        
        value = asarray(value, self._typecode)
        if value.shape == (1,1):
            value = value[0,0]
        self.array[ind] = value
        
    def __delitem__(self,ind):
        """
        Deletes column with index ind.
        """
        self.del_row(ind)
        
    #def put(self,inds,subvector):
    #    put(self.array,inds,subvector.array)

    #def take(self,inds):
    #    return self._rc(take(self.array,inds))

    def put_array(self,inds,subarray):
        put(self.array,inds,subarray)

    def take_array(self,inds):
        return take(self.array,inds)

    def put_row(self,rowind,array):
        put(self.array,rowind,array)
        
    def put_rowarray(self,rowind,inds,array):
        put(self.array[rowind],inds,array)

    def take_rowarray(self,rowind,inds):
        return take(self.array[rowind],inds)

    def __getstate__(self):
        return {'array':self.array}
    
    def __setstate__(self,state):
        self.array=state['array']
        self.shape = self.array.shape
        self._typecode = self.array.typecode()
        
        
class Array_db:
    """
    Class which administrates a multi-array object, where each array of
    the multi-array object can be addressed by a keyword, similar to a
    dictionary.

    """
    def __init__(self):
        self.db={}
        self.arrays=array([])

    def __call__(self):
        return self.arrays

        
    def astype(self, typecode): return self._rc(self.array.astype(typecode))
    
    def replace(self,key,array):
        self.__delitem__(key)
        self.add(key,array)
        return self.db[key][0]
            
    def add(self,key,array):
        add_len=len(array)
        if self.db.has_key(key):
            array=concatenate((self.arrays[self.db[key][0]:self.db[key][1]+1],
                               array)
                              )
            #print 'begin add--------------------------------------'
            #print 'array',array
            #print 'before del'
            #print 'self.db',self.db
            #print 'self.arrays',self.arrays
            self.__delitem__(key)
            #print 'after del'
            #print 'self.db',self.db
            #print 'self.arrays',self.arrays
        if len(self.db)==0:
            #print 'index interval (init)',0,len(array)-1
            self.arrays=array
            self.db[key]=(0,len(array)-1)
        else:
            #print 'index interval',len(self.arrays),len(self.arrays)+len(array)-1
            self.db[key]=(len(self.arrays),len(self.arrays)+len(array)-1)
            self.arrays=concatenate((self.arrays,array))
        #print 'return ind:',self.db[key][1]-len(array)+1

        #print 'after add==============================='
        #print 'self.db',self.db
        #print 'self.arrays',self.arrays
        return self.db[key][1]-add_len+1

    
            

    def __setitem__(self,key,a):
        if self.db.has_key(key):
            self.arrays[self.db[key][0]:self.db[key][1]+1]=a
        else:
            self.add(key,array([a]))
        #return self.db[key][1]-len(array)+1
    
    def __getitem__(self,key):
        #print 'db=\n',self.db
        return self.arrays[self.db[key][0]:self.db[key][1]+1]

    def __delitem__(self,key):
        if self.db.has_key(key):
            ind_first, ind_last=self.db[key]
            ind_len=len(self.__getitem__(key))
            #print 'delete...'
            #print 'index=',self.db[key]
            #print 'delete item =\n',self.db[key]
            #print 'ind_len=',ind_len
            if ind_first!=0:
                self.arrays=concatenate((
                    self.arrays[0:ind_first],
                    self.arrays[ind_last+1:len(self.arrays)]
                    ))
            else:
                self.arrays=self.arrays[ind_last+1:len(self.arrays)]
            del self.db[key]
            for k in self.db.keys():
                if self.db[k][0]>ind_last:
                    self.db[k]=(self.db[k][0]-ind_len,
                                self.db[k][1]-ind_len)
    def get_key_with_index(self,ind):
        '''Get key for array containing index ind'''
        if len(self.db.keys())==1: # exception if there is only one entry
            return self.db.keys()[0]
        else:
            for key in self.db.keys():
                fromind,toind=self.db[key]
                if (ind>=fromind) & (ind<=toind):
                    return key
            return None

    def del_index(self,ind):
        #obj=self.objs[ind]
        for key in self.db.keys():
            fromind,toind=self.db[key]
            if (ind>=fromind) & (ind<=toind):
                remarrays=1*self.__getitem__(key)
                self.__delitem__(key)
                remind=ind-fromind
                #print 'key',key
                #print 'remind',remind
                #print 'del index',ind-fromind
                #print 'remarrays',remarrays
                #print 'self.arrays',self.arrays
                #print 'self.db',self.db

                if remind!=0:
                    remarrays=concatenate((
                        remarrays[0:remind],
                        remarrays[remind+1:len(remarrays)]))
                else:
                    remarrays=remarrays[1:len(remarrays)]
                #print 'remarrays after del',remarrays
                if len(remarrays)>0:
                    self.add(key,remarrays)
                break
            
    def take(self,ind_list): return take(self.arrays,ind_list)
    
    def multiply_with_matrix(self,matrix):
        self.arrays=transpose(matrixmultiply(matrix,self.arrays),(1,0,2))
            
class Obj_db:
    """
    Same as Array_db, but with a list with python objects instead of arrays.

    """
    def __init__(self):
        self.db={}
        self.objs=[]
        
    def replace(self,key,string):
        self.__delitem__(key)
        self.add(key,string)
        return self.db[key][0]

    
    def add(self,key,objlist):
        add_len=len(objlist)
        if self.db.has_key(key):
            objlist=self.objs[self.db[key][0]:self.db[key][1]+1]\
                         +objlist
            self.__delitem__(key)

        if len(self.db)==0:
            #print 'index interval (init)',0,len(string)-1
            self.objs=objlist
            self.db[key]=(0,len(objlist)-1)
        else:
            #print 'index interval',len(self.objs),len(self.objs)+len(string)-1
            self.db[key]=(len(self.objs),len(self.objs)\
                          +len(objlist)-1)
            self.objs=self.objs+objlist
        #print 'return ind:',self.db[key][1]-len(string)+1


        return self.db[key][1]-add_len+1

    def objs_to_ids(self,sim):
        """
        This turns the objects in oth internal object list to ids,
        used by the current simulation sim.
        """
        self.objs=sim.get_ids(self.objs)

    def ids_to_objs(self,sim):
        """
        This turns the object ids of internal object list to object pointers,
        used by the current simulation sim.
        """
        self.objs=sim.rebuild_pointers(self.objs)
        
        
    def __setitem__(self,key,objlist):
        self.objs[self.db[key][0]:self.db[key][1]+1]=objlist

    
    def __getitem__(self,key):
        #print 'db=\n',self.db
        return self.objs[self.db[key][0]:self.db[key][1]+1]

    def __delitem__(self,key):
        if self.db.has_key(key):
            ind_first, ind_last=self.db[key]
            ind_len=len(self.__getitem__(key))
            if ind_first!=0:
                self.objs=self.objs[0:ind_first]+\
                              self.objs[ind_last+1:len(self.objs)]
            else:
                self.objs=self.objs[ind_last+1:len(self.objs)]
            del self.db[key]
            for k in self.db.keys():
                if self.db[k][0]>ind_last:
                    self.db[k]=(self.db[k][0]-ind_len,
                                self.db[k][1]-ind_len)

    def get_index(self,obj):
        return self.objs.index(obj)

    def del_index(self,ind):
        for key in self.db.keys():
            fromind,toind=self.db[key]
            if (ind>=fromind) & (ind<=toind):
                remobjs=1*self.__getitem__(key)
                self.__delitem__(key)
                #print 'key',key
                #print 'del index',ind-fromind
                #print 'remobjs',remobjs
                #print 'self.objs',self.objs
                #print 'self.db',self.db
                
                del remobjs[ind-fromind]
                if len(remobjs)>0:
                    self.add(key,remobjs)
                break
            
    def del_obj(self,obj):
        self.del_index(self.get_index(obj))
        
###############################################################################
if __name__ == '__main__':
    a=Array_db()
    ind=1
    a[ind]=1.1
    print 'key of index',ind,'=',a.get_key_with_index(ind)
    
