# python its.py
__version__='0.47'

copyright="""
  Copyrigh statement for iTS MAIT version """+__version__+\
  """
  
  This software comes with absolutely no guarantee.
  MAIT international e.V. takes no responsibility for the results
  that the software produces, neither for any direct or indirect
  damage it may cause.

  This version can only be used, copied or distributed with a written
  permission of the owner, MAIT international e.V., Germany.
  E-mail: info@MAITint.org

  Customized PRT/AGV or GRT system simulation modules as well as
  modules for special analyses can be developed on demand.  """

usage="""
USAGE:
    
    With the ''iTS MAIT software'' you can create, save, open and simulate
    innovative transportation networks.
    
    python its.py [options] [<filename>]
    
    When called without any optional arguments, the graphical
    innovative transport network editor will be invoked. When called only
    with the <filename>, then the respective simulation will be
    opened into the graphical editor-window.
    

    -c or --cml   command line mode, runs simulation on <simfilename>
                  without windows or support for graphical interface.

    -e            sets end-of-simulation time in seconds, which means that
                  the simulation in <filename> will terminate at this time.
                  Intended to be used with the --cml option.

    -d            sets the duration of the simulation in seconds,
                  which means that
                  the simulation in <filename> will be simulated for
                  this time, starting with the present time.
                  Intended to be used with the --cml option.

    -o <outputfilename> or --output <outputfilename>
                  at the end of the simulation, simulation 
                  is saved to <outputfilename>.
                  Intended to be used with the --cml option.

    -f or --feedback
                  the output simulation file will be fed back into
                  <filename> , i.e <outputfilename>=<filename>
                  Intended to be used with the --cml option.
                  
    -s            sets the sampling time, used for iterations.

    -r <reportfilename> or --report <reportfilename>
                  a comma separated, table-stype report will be saved in
                  <reportfilename> at the end of the simulation.
                  Intended to be used with the --cml option.

    -h or --help  shows this message.
    """


###############################################################################
# IMPORT

import sys
sys.path[:0] = ['classlib']
#sys.path.append('classlib/gui')
import Tkinter,tkMessageBox,tkFileDialog,Pmw,gui
import getopt
import time, os, string, glob, pickle,types
from Numeric import zeros,ones
#import zipfile
###############################################################################
# CONFIG (should be eliminated)

debug=3
type_sep='.'
id_sep='/'
part_sep=':'
node_attract=10
###############################################################################
# UTILITY FUNCTIONS


###############################################################################
# CLASS DEFINITIONS


class Sim:
    """
    Simulation class application.
    """
    #def __init__(self,simfile_name,param,simdir='',
    #             simwin=None,menubar=None,balloon=None,messagebar=None):
    def __init__(self,simfile_name,**args):
        # set default version....will be done in set_param
        # self.__version__=param['version']
        self.__version__=__version__

        print """
    This is the innovative Transport Simulator by MAIT international
                       iTS-MAIT version""",__version__,""".\n"""

        print copyright+'\n'

        # determine parameters dependent if they where passed by
        # key-value pairs or by a single parameter dictionary
        #print 'args=',args
        if args.has_key('params'):
            param=args['params']
        else:
            param=args
        
        # this is to indicate the tye og application
        self.app_type='sim'
        
        # set operation mode, mode key is required
        self.mode=param.get('mode','')

        
        # default directory with simulation-projects
        self.simdir=args.get('simdir','')

        # a simulation, once started will continue
        self.continue_flag=1

        # pre-initializialise variables:
        self.realtime_start=time.time()
        self.time_comp=0.0 # will be set to time_end when simulation is started
        # Default simulation parameters.
        # They will be overwritten in init_sim
        # first by the parameters set in the
        # simulation file and then by the parameters from the command line
        # options.
        self._param={}
        
        self.set_simfile_name(simfile_name)
        self.time_start=0.0
        self.id_sep=id_sep
        self.type_sep=type_sep
        self.part_sep=part_sep
        self.node_attract=node_attract
        self.objectlibbrowser=None
        self.logtasks={}
        

        # define gui stuff
        self.simwin=args.get('simwin',None)
        self.menubar=args.get('menubar',None)
        self.balloon=args.get('balloon',None)
        self.messagebar=args.get('messagebar',None)

        
        # this flag will be set during simulation is running
        self.sim_flag=0

        # object library path
        self.objlib_path=os.path.abspath(sys.path[1])+os.sep+'objlib'
        
        # global table that contains all simulation objects
        self.objs={}      # objects database with obj.type as key

        # global library db, one for each simulation object
        self.libs={} # with type as key as in objs db
        
        self.maincycle_flag=0 # =1 during  main-cycle
        
        # objects whose iterate method is called during the mainloop
        self.objs_iterate={} # objects database with obj.type as key
        
        
        # objects whose output method is called during the mainloop
        self.objs_output={} # objects database with obj.type as key
        self.output_flag=0 # =1 during output in main-cycle

        
        
        # call  loop dependent on simulation mode
        if self.mode=="gui":

            # pointer to panel which will be used to set sim parameters
            self.simpanel=None
            
            # this is the rate with which all the widgets are updated
            # 500 means each 500 main cycles a widged update is triggered
            self.gui_update_rate=500
  
            #if self.simfile_name:
                # remove empty window widget in the main simulation window
                #self.simwin.empty.destroy()
            
            self.time=0.0
            self.messagebar.message('state','Simulation stopped at '+\
                                    self.seconds_to_string(self.time))

            self.menubar.addmenu('Edit',
                                 'Edit transport network and other objects',
                                  tearoff=0)

            self.menubar.addmenu('View',
                                 'Change various properities of graphical '+\
                                 'representation',
                                  tearoff=1)
            
            # create menu with callback functions
            self.menubar.addmenu('Simulation',
                                 'Control and configure simulation',
                                 tearoff=1)

            self.menubar.addmenuitem('Simulation', 'command',\
                          'Start simulation, reset time and statistics',
                                         command = self.start_gui,
                                         label = 'Start')
            
            self.menubar.addmenuitem('Simulation', 'command',\
                                         'Stop simulation',
                                         command = self.stop_gui,
                                         label = 'Stop')

            self.menubar.addmenuitem('Simulation', 'command',\
                                         'Continue simulation',
                                         command = self.continue_gui,
                                         label = 'Continue')
            
            self.menubar.addmenuitem('Simulation', 'separator')
            
            self.menubar.addmenuitem('Simulation', 'command',\
                                   'Set and display  simulation parameters',
                                         command = self.set_param_gui,
                                         label = 'Set parameters...')
            

            self.menubar.addmenu('Tools',
                        'Tools to analyze the current state of the simulation',
                                     tearoff=0)
            self.menubar.addmenuitem('Tools', 'command',\
                                   'Exports results of current simulation'+\
                                     ' to spreadsheet text-file',
                                     command = self.export_results,
                                     label = 'Export results...')
            
            self.menubar.addmenuitem('Tools', 'command',\
                                   'Exports costs of current simulation'+\
                                     ' to spreadsheet text-file',
                                     command = self.export_costs,
                                     label = 'Export costs...')
            
                                     
            self.menubar.addmenuitem('Tools', 'command',\
                                   'Export all current stops'+\
                                     ' to spreadsheet text-file',
                                     command = self.export_stops,
                                     label = 'Export stops...')
            
            self.menubar.addmenuitem('Tools', 'command',\
                                     'Export waiting time statistics for stops'+\
                                     ' to text-file',
                                     command = self.export_waittimes,
                                     label = 'Export waiting times...')
                                     
            
            
            # Initialise the checkbutton for show_ghosts to 0:
            self.var_log_waiting_users= Tkinter.IntVar()
            self.var_log_waiting_users.set(0)                      
            self.menubar.addmenuitem('Tools', 'checkbutton',
                                    'Log number of waiting users at stops into a'+\
                                     'text-file',
                                    command = self._toggle_log_waiting_users,
                                    variable = self.var_log_waiting_users,
                                    label = 'Log waiting users...')
                                                                 
            self.menubar.addmenuitem('Tools', 'command',\
                                   'Export triptime matrix...',
                                     command = self.export_stops_triptimes,
                                     label = 'Export triptime matrix....')
            
            
            
            self.menubar.addmenuitem('Tools', 'command',\
                                   'Generate template ODM...',
                                     command = self.odgen_template,
                                     label = 'Generate template ODM...')
            
            self.init_sim(param) # initialize simulation objects

            # simulate a setp when pressing enter
            self.simwin.bind('<Return>',self.sim_step_gui)
            
            # the gui initialization ends here and waits untill
            # Simulation>Start is chosen from the menue
            
        elif self.mode=="cml":
            self.init_sim(param) # initialize simulation objects
            self.loop_cml()
        else:
            self.init_sim(param) # initialize simulation objects
            # no loop executed, wait for explicit call of start_sim




    def _toggle_log_waiting_users(self,filename=''):
        """
        Log number of waiting users at stops into a text file.
        """
        if self.var_log_waiting_users.get():
            sep='\t'
    
            
            if (not filename)&(self.mode=='gui'):
                if self.sim_flag:
                    self.stop_gui()
                filename=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                              filetypes=[('Text files','*.txt'),\
                                                         ('All files','*')]).show()
                if not filename: return
    
            elif not filename:
                filename=self.simdir+os.sep+\
                     os.path.splitext(os.path.basename(self.simfilename))[0]+\
                                           '_stop_triptimes'+'.txt'
    
            
            self.init_log_waiting_users(filename,sep)
            #print text
            
        else:
            # remove logging from task list
            (log_waiting_users,logfile,closelogfunction)=self.logtasks['waiting_users']
            closelogfunction(logfile)
           
    def init_log_waiting_users(self,filename,sep='\t',time_log=10.0):
        """
        Initial logfile for waiting_users.
        Can be used by scripts
        """
        self.time_log=time_log
        self.time_nextlog=self.time
        #print 'init_log_waiting_users',filename,self.time_log
        
        logfile=open(filename,'w')
        for row in self.get_exportfileheader(sep,name='Number of waiting users at stops'):
            logfile.write(row+'\n')
            
        # create a sorted list with names of stops
        stop_names=self.get_stops().keys()
        #for stop in self.get_stops(all=1).values():
        #    stop_names.append((stop.name,stop.id))
        
        stop_names.sort()
        
        row='time [hh:mm:ss]'+sep
        for name in stop_names: row+=name+sep
        logfile.write(row+'\n')
        
        
        # append logtask to loglist
        self.logtasks['waiting_users']=(self.log_waiting_users,
                                        logfile,
                                        self.close_log_waiting_users)
        
    def log_waiting_users(self,logfile,sep='\t'):
        if self.time>self.time_nextlog:
            self.time_nextlog+=self.time_log
            #print 'log_waiting_users',self.seconds_to_string(self.time)
            
            # create a sorted list with names of stops
            stop_names=[]
            stops=self.get_stops()
            stop_names=stops.keys()
            stop_names.sort()
            
            row=self.seconds_to_string(self.time)+sep
            for name in stop_names: 
                row+=`stops[name].n_users_queuing`+sep
                
            logfile.write(row+'\n')
      
        
    
    def close_log_waiting_users(self,logfile):
        self.var_log_waiting_users.set(0)     
        logfile.close()
        del self.logtasks['waiting_users']
            
    def log_data(self):
        
        #print 'log_data at',self.time
        
        for logfunction,logfile,closelogfunction in self.logtasks.values():
            logfunction(logfile)
            
    def close_log_data(self):
        """
        Closes all log files (currently not in use.)
        """
        
        print 'close_log_data at',self.seconds_to_string(self.time)
        
        for logfunction,logfile ,closelogfunction in self.logtasks.values():
            closelogfunction(logfile)
            
            
    def export_stops_triptimes(self,filename=''):
        """
        Exports data of current stops with average triptimes to
        spreadsheet formatted text file
        """
        sep='\t'

        
        if (not filename)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            filename=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()
            if not filename: return

        elif not filename:
            filename=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfilename))[0]+\
                                       '_stop_triptimes'+'.txt'

        # create a sorted list with names of stops
        stop_names=[]
        for stop in self.get_stops(all=1).values():
            stop_names.append((stop.name,stop.id))
        stop_names.sort()

        # create a dict with id as key and index of stop_names as value
        stop_ind={}
        ind=0
        for (name,id) in stop_names:
            stop_ind[id]=ind
            ind+=1

        # define array
        n_stops=len(stop_ind)
        triptimes=zeros((n_stops,n_stops),typecode='float')
        # this counts the number of trips for each OD pair
        # -1 means that there are no trips 
        tripnumbers=ones((n_stops,n_stops),typecode='int')
        # go through all users types
        for obj in self.objs.values():
            #print ' scan',obj.id,self.split_id(obj.id)['category']
            if self.split_id(obj.id)['category']=='user':
            #if obj.id=='base.user.generic':
                for user in obj.mods.values():
                    #print user.id,user.trips_total,user.orig_id,user.dest_id
                    if user.trips_total==1:
                        if stop_ind.has_key(user.dest_id)\
                           & stop_ind.has_key(user.orig_id):
                            ind_orig=stop_ind[user.orig_id]
                            ind_dest=stop_ind[user.dest_id]

                            # increse trip time for thid OD pair
                            triptimes[ind_orig,ind_dest]=triptimes[
                                ind_orig,ind_dest]+user.triptime_total

                            # increase also the counter later used for average
                            if tripnumbers[ind_orig,ind_dest]<0.0:
                                # initialize
                                tripnumbers[ind_orig,ind_dest]=1.0
                            else:
                                # increase
                                tripnumbers[ind_orig,ind_dest]=tripnumbers[
                                    ind_orig,ind_dest]+1.0
        # now average
        triptimes_av=triptimes/tripnumbers

        # write header line
        row=sep
        for (name,id) in stop_names:
            row+=(name+sep)
        text=[row]
        # do matrix
        for ind_orig in range(n_stops):
            (row,id) = stop_names[ind_orig]
            for ind_dest in range(n_stops):
                data=triptimes_av[ind_orig,ind_dest]
                row+=sep+str(data)
            text.append(row)
            
        #print text
        f=open(filename,'w')
        for row in self.get_exportfileheader(sep,name='Triptimes between stops in seconds'):
            f.write(row+'\n')
        for row in text:
            f.write(row+'\n')
        f.close()

    def export_zones_triptimes(self,filename=''):
        """
        Exports data of current zones with average triptimes to
        spreadsheet formatted text file
        """
        sep='\t'

        
        if (not filename)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            filename=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()
            if not filename: return

        elif not filename:
            filename=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfilename))[0]+\
                                       '_stop_triptimes'+'.txt'

        # create a sorted list with names of zones
        stop_names=[]
        stopzones={}
        stops=self.get_stops(all=1).values()
        for stop in stops:
            stop_names.append((stop.name,stop.id))
            if not stopzones.has_key(stop.zone):
                stopzones[stop.zone]=[]
            stopzones[stop.zone].append(stop.id)

        name_stopzones=stopzones.keys()
        name_stopzones.sort()

        
            
          
        # create a dict with stop.id as key and index of
        # stopzone name list as value
        stopzone_ind={}
        ind=0
        for nz in name_stopzones:
            for id in stopzones[nz]:
                stopzone_ind[id]=ind
            ind+=1

        # define array
        n=len(stopzones)
        triptimes=zeros((n,n),typecode='float')
        # this counts the number of trips for each OD pair
        # -1 means that there are no trips 
        tripnumbers=ones((n,n),typecode='int')
        # go through all users types
        for obj in self.objs.values():
            if self.split_id(obj.id)['category']=='user':
            #if obj.id=='base.user.generic':
                for user in obj.mods.values():
                    if user.trips_total==1:
                        if stopzone_ind.has_key(user.dest_id)\
                           & stopzone_ind.has_key(user.orig_id):
                            ind_orig=stopzone_ind[user.orig_id]
                            ind_dest=stopzone_ind[user.dest_id]

                            # increse trip time for thid OD pair
                            triptimes[ind_orig,ind_dest]=triptimes[
                                ind_orig,ind_dest]+user.triptime_total

                            # increase also the counter later used for average
                            if tripnumbers[ind_orig,ind_dest]<0.0:
                                # initialize
                                tripnumbers[ind_orig,ind_dest]=1.0
                            else:
                                # increase
                                tripnumbers[ind_orig,ind_dest]=tripnumbers[
                                    ind_orig,ind_dest]+1.0
        # now average
        triptimes_av=triptimes/tripnumbers

        # write header line
        row=sep
        for name in name_stopzones:
            row+=(name+sep)
        text=[row]

        # do matrix
	for ind_orig in range(n):
                row = name_stopzones[ind_orig]
		for ind_dest in range(n):
			data=triptimes_av[ind_orig,ind_dest]
			row+=sep+str(data)
		text.append(row)
		
        #print text
        f=open(filename,'w')
        for row in self.get_exportfileheader(sep):
            f.write(row+'\n')
        for row in text:
            f.write(row+'\n')
        f.close()
     
    def invoke_odgen(self):
        """
        invoke OD generator utility.
        """
        self.append_obj('base.net.odgen')

    def odgen_template(self):
        """
        Generates template for  station to station OD matrices 
        """
        filename =tkFileDialog.SaveAs(initialdir=self.simdir,
                                      filetypes=[('Text files','*.txt'),
                                                 ('All files','*')]).show()

        stops={}
        for stop in self.get_stops().values():
            stops[stop.name]=stop
            
        
        sep=','   # column separator of od file

        # make header line
        header=stops.keys()
        header.sort()

        text=[' ']
        for name in header:
            text[0]+=sep+name
        
        #print text
        
        # generate matrix
        for name_row in header:
            row=name_row
            for name_col in header:
                if name_col!=name_row:
                    row+=sep+str(stops[name_row].berthssec.berth_no)
                else:
                    row+=sep+'0'
            text.append(row)

        #print text
        f=open(filename,'w')
        for row in text:
            f.write(row+'\n')
        f.close()
                

        

    def set_simfile_name(self,set_simfile_name):
        """
        Sets the name of this simulation.
        Attention this name contains the entire directory path.
        """
        #??????????? projects/xxx.its as simfile_name ????
        self.simfile_dirname=os.path.dirname(set_simfile_name)
        self.simfile_name=set_simfile_name
        
    def set_param(self,param):
        '''
        Sets global simulation parameters
        This is a very messy function because no individual paramester
        can be set, without resetting all others!!!!
        '''
  
        # read out parameters and store them in globally accessible variables
        #self.simfile_name=param['simfile_name']
        #self.__version__=param.get('version')
        
        #print 'set_param',param
        self.debug=param.get('debug',3)

        if self.debug>=5:
            print 'Version:',self.__version__


        self.time_sample=param.get('time_sample',0.01)
        self.time=param.get('time',0)
        if param.has_key('time_delta'):
            self.time_end=self.time+param['time_delta']
            # remove this parameter as it has done its job
            del param['time_delta']
        else:
            self.time_end=param.get('time_end',3600.0)

        self.currency=param.get('currency','EUR')
        self.filename_output=param.get('filename_output','')
        self.filename_report=param.get('filename_report','')
        self.stop_name_count=param.get('stop_name_count',0)

        # updating with all explicitely given params
        self._param.update(param)

        # updating with possibly modified or newly initialized papameters
        self._param['debug']=self.debug
        self._param['time_sample']=self.time_sample
        self._param['currency']=self.currency
        self._param['time_end']=self.time_end
        self._param['stop_name_count']=self.stop_name_count
        self._param['filename_output']=self.filename_output
        self._param['filename_report']=self.filename_report
        
    def get_param(self,key=None):
        '''Returns global simulation parameter for given key'''
  
        # read out parameters and store them in globally accessible variables
        self._param['version']=self.__version__
        self._param['time']=self.time
        self._param['stop_name_count']=self.stop_name_count
        if key:
            if self._param.has_key(key):
                return self._param[key]
            else:
                return None
        else:
            return self._param.copy()
        
      

    def close(self):
        """
        Does necessary action before this simulation is
        erased from memory.
        """
        self.close_log_data()
        
        if self.objectlibbrowser:
            self.objectlibbrowser.close()
            del self.objectlibbrowser

        for obj in self.objs.values(): obj.close()

        self.menubar.deletemenu('Edit')
        self.menubar.deletemenu('Simulation')
        self.menubar.deletemenu('Tools')

        
    def stop(self):
        """
        Stops current simulation, without resetting anything.
        """
        self.continue_flag=0
        
    def stop_gui(self):
        """
        Stops simulation.
        """
        self.stop()

    def continue_gui(self):
        """
        Continues  simulation untill time_end.
        """
        if not self.sim_flag:
            self.messagebar.message('state','Simulation running')

            self.prepare_continue()
        
            # set the speed how often gui objects will be updated
            self.gui_update_count=self.gui_update_rate

            # go...
            self.loop_gui()

    def prepare_continue(self):
        """
        This function prepares simulation objects and environment
        to continue the simulation.
        Hereafter, it is up to the calling function to invokr the
        maincycle method.
        """
        self.time_comp=self.time_end
        self.time_start=self.time
        self.realtime_start=time.time() # save start time in seconds
        self.sim_flag=self.time < self.time_comp
        self.continue_flag=1
        
    def get_stop_name(self):
        """
        Returns a string with a unique stop name.
        This is a function that will be provided by the net management.
        """
        self.stop_name_count+=1
        return str(self.stop_name_count)
    
    def get_stops(self,all=0):
        """
        Returns a database with the currently availlable (active) stops
        with stop name as key and the stop module as value.
        If all=1 also non-active stops will be returned.
        """
        stops={}
        stops_all=[]
        for id in self.objs.keys():
            id_data=self.split_id(id)
            if (id_data['category']=='track'):
                if string.find(id_data['name'],'stop')!= -1:
                    stops_all=id_data['obj'].mods.values()

                    # append only active stops
                    for stop in stops_all:
                        if (stop.status=='active')&(not all):
                            #print 'get_stops:',stop.name,stop.id
                            stops[stop.name]=stop
                        elif all:
                            #print 'get_stops:',stop.name,stop.id
                            stops[stop.name]=stop
        return stops
    
    
    
    def export_waittimes(self,file_name=''):
        """
        Exports waittime info of all current stops to
        spreadsheet formatted text file
        """
        keylist=['name','waitingtime_users_queued_av','waitingtime_min',
                'waitingtime_max','waitingtime_dev','n_loads_per_h']
        
        descriptors_extra=[{'key':'waitingtime_av_weighted',
                            'name':'Weighted Avg Wait',
                            #'name':'Average waiting time',
                            'type':'real',
                            'printformat':'%.2f',
                            'units':{'s':1},
                            'width':1,
                            'valid':{'min':0,'max':10**8},
                            'mode':'read',
                            }]
        for descriptor in descriptors_extra:
            keylist.append(descriptor['key'])
        
        sep='\t'
        #sep=','

        
        if (not file_name)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            file_name=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()
            if not file_name: return

        elif not file_name:
            file_name=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfile_name))[0]+\
                                       '_stops'+'.txt'

        
                            
        dataset_sum={}
        for key in keylist:
            dataset_sum[key]=0.0
        
        # do precalcs and summerizing statistics
        n_loads_per_h_total=0
        n_stops=0
        for stop in self.get_stops(all=1).values():
            descriptors=[];dataset={}
            stop.format_statistics_specific(descriptors,dataset)
            n_loads_per_h_total+=dataset['n_loads_per_h']
            #print '  stop',stop.id,stop.zone,n_stops
            if stop.zone!='maintenance':
                n_stops+=1
        
        report=self.get_exportfileheader(sep=sep,name='Waiting time info of stops')
           
        is_header_written=False
        #print 'stops:',self.get_stops(all=1).keys()
        for stop in self.get_stops(all=1).values():
            descriptors=[];dataset={}
            stop.format_statistics_specific(descriptors,dataset)
            descriptors+=descriptors_extra
            #for descriptor in descriptors:
            #    print '=====descriptor key:',descriptor['key']
            #print 'keylist=',keylist
            if n_loads_per_h_total>0:
                dataset['waitingtime_av_weighted']=dataset['waitingtime_users_queued_av']\
                    *dataset['n_loads_per_h'] /n_loads_per_h_total
            else:
                dataset['waitingtime_av_weighted']=0.0
            
            if not is_header_written:
                line=''
                for descriptor in descriptors:
                    if descriptor['key'] in keylist:
                        if descriptor.has_key('units'):
                            line+=descriptor['name'] \
                                +' ['+descriptor['units'].keys()[0]+']'+sep
                        else:
                             line+=descriptor['name']+sep
                             
                report.append(line)
                #line='-'*len(line)
                #line='\n'
                #report.append(line)
                is_header_written=True
            
            # now write data columns to line
            line=''
            for descriptor in descriptors:
                
                if (descriptor['key'] in keylist)&(stop.zone!='maintenance'):
                    key=descriptor['key']
                    value=dataset[key]
                    #print stop.name,key,dataset[key]
                    if descriptor.has_key('printformat'):
                        line+=descriptor['printformat']%value
                    else:
                        line+=str(value)
                    
                    line+=sep
                    if key!='name':
                        dataset_sum[key]+=value
                
            report.append(line)
            
        report.append('\n')
        
        line='Simple Total Averages'+sep
        line_minutes='Minutes'+sep
        for descriptor in descriptors:
                if (descriptor['key'] in keylist)&(descriptor['key']!='name'):
                    key=descriptor['key']
                    value=dataset_sum[key]/n_stops
                    #print stop.name,key,dataset[key]
                    if descriptor.has_key('printformat'):
                        line+=descriptor['printformat']%value
                    else:
                        line+=str(value)
                    if descriptor.has_key('units'):
                        if descriptor['units'].keys()[0]=='s':
                            #print '  format time:'
                            line_minutes+=self.seconds_to_string(value)
                    
                    
                    line+=sep
                    line_minutes+=sep
        
        report.append(line)       
        report.append(line_minutes+'\n')
        report.append('Total Loads/h'+sep +`'%.2f'%dataset_sum['n_loads_per_h']`+'\n')
                        
        report.append('Weighted Total Avg Wait [s]'+sep\
                        +'%.2f'%(dataset_sum['waitingtime_av_weighted']))
                
        
        # get this total average waiting times from report
        userman=self.objs['base.userman.experim']
        userman.format_statistics_specific(descriptors,dataset_sum)
        
        report.append('Result Rpt - Avg Wait [s]'+sep\
                       +'%.2f'%(dataset_sum['Average waiting times']) )
        
        
        # write to file
        file=open(file_name,'w')
        for line in report:
            #print line
            file.write(line+'\n')
            #print line
        file.close()
     
    def get_current_date_time(self,dsep='/',hsep=':'):
        t=time.localtime()
        return `t[0]`+dsep+`t[1]`+dsep+`t[2]`\
                +'  '+`t[3]`+hsep+`t[4]`+hsep+`t[5]`
        
    def get_exportfileheader(self,sep='\t',dsep='/',hsep=':',name=''):
        if name:
            stringlist=[name+'\n']
        else:
            stringlist=[]
            
        stringlist.append("iTS MAIT version "+sep+__version__)
        stringlist.append('Simulation-file'+sep+self.simfile_name+'\n')
        stringlist.append('Creation date:'+sep+self.get_current_date_time(dsep,hsep))
        
        if self.time-self.time_start>1.0:
            stringlist.append('Simulation duration:'+sep\
                                +self.seconds_to_string(self.time-self.time_start)\
                                +' [hh:mm:ss]')
        stringlist.append('\n')
        
        return stringlist
        
    def export_costs(self,file_name=''):
        """
        Exports waittime info of all current stops to
        spreadsheet formatted text file
        """
        keylist=['name','waitingtime_users_queued_av','waitingtime_min',
                'waitingtime_max','waitingtime_dev','n_loads_per_h']
        sep='\t'
        #sep=','

        
        if (not file_name)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            file_name=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()
            if not file_name: return

        elif not file_name:
            file_name=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfile_name))[0]+\
                                       '_stops'+'.txt'
    
        # define all columns in the cost table
        
        descriptors=[
                {'name':'Type',
               'key':'type',
               'type':'text',
               },
                {'name':'Units',
               'key':'no',
               'type':'int',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               },
               {'name':'Costs/unit',
               'key':'estim_init_pu',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
               {'name':'Costs/unit/a',
               'key':'estim_pa_pu',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency+'/a':1},
               },
               {'name':'Total length',
               'key':'length_total',
               'type':'real',
               #'printformat':'%.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{'m':1},
               },
               {'name':'Costs/meter',
               'key':'estim_init_pm',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency+'/m':1},
               },
               {'name':'Costs/meter/a',
               'key':'estim_pa_pm',
               'type':'real',
               'printformat':'%.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency+'/m/a':1},
               },
               {'name':'Invest. costs',
               'key':'estim_init',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
              {'name':'O&M costs/a',
               'key':'estim_pa',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
               {'name':'Simulated Revenue',
               'key':'revenue',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
              ]

                  
        # initialize report (list of strings)
        stringlist=self.get_exportfileheader(sep,name='System costs table')
        
        # make header
        line=''
        for descriptor in descriptors:
            # column names
            line+=descriptor['name']+sep
        stringlist.append(line)
        
        line=''
        for descriptor in descriptors:
            # units
            if descriptor.has_key('units'):
                line+=descriptor['units'].keys()[0]+sep
            else:
                line+=sep
        stringlist.append(line+'\n')
            

        # generate an ordered list for reports
        ids_ordered=self.objs.keys()
        ids_ordered.sort()
        
        length_total_track=0.0
        for id in ids_ordered:
            obj=self.objs[id]
            category=self.split_id(obj.id)['category']
            #print '\n\n'+string.upper(obj.name)+'\n'
            # no net objects in list...not relevant, only confusing
            costs= {'type':obj.type}
            if category not in ('net','user'):
                
                #print '\n\n***********  type=',obj.type
                
                if obj.multimod_flag:
                    # do some pre-calcs for objects containing modules
                    no=len(obj.mods)
                    costs['no']=no
                    if no>0:
                        mod=obj.mods.values()[0]
                        mod_descriptor,mod_costs=mod.format_costs()
                        
                        #print '  mod_costs=',mod_costs
                        
                        if (category=='track')&(len(mod_costs)>0):
                            length=0.0
                            for testmod in obj.mods.values():
                                length+=testmod.get_length()
                            #length*=1000 # convert in mm
                            
                            #print '      length=',length
                            length_total_track+=length
                            costs['length_total']=length
                            costs['estim_init_pu']=mod_costs['estim_init_ex']
                            costs['estim_pa_pu']=mod_costs['estim_pa_ex']
                            
                            costs['estim_init_pm']=mod_costs['estim_init_pm']
                            costs['estim_pa_pm']=mod_costs['estim_pa_pm']
                            
                            costs['estim_init']=costs['estim_init_pm']*length \
                                                +costs['estim_init_pu']*no
                            costs['estim_pa']=costs['estim_pa_pm']*length \
                                                +costs['estim_pa_pu']*no
                                
                        elif (len(mod_costs)>0):
                            costs['estim_init_pu']=mod_costs['estim_init']
                            costs['estim_pa_pu']=mod_costs['estim_pa']
                            
                            costs['estim_init']=costs['estim_init_pu']*no
                            costs['estim_pa']=costs['estim_pa_pu']*no
                            
                            
                            
                else:
                    # costs for single unit object
                    obj_descriptor,obj_costs=obj.report_costs()
                    costs['no']=1
                    costs['estim_init_pu']=obj_costs['estim_init']
                    costs['estim_pa_pu']=obj_costs['estim_pa']
                    costs['estim_init']=obj_costs['estim_init']
                    costs['estim_pa']=obj_costs['estim_pa']
                    if obj_costs.has_key('revenue'):
                        costs['revenue']=obj_costs['revenue']
                
                #print ' now compose line with costs=',costs
                line=''
                for descriptor in descriptors:
                    key=descriptor['key']
                    data=''
                    if costs.has_key(key):
                        #print stop.name,key,dataset[key]
                        data=costs[key]
                        
                    # conclude column    
                    if (descriptor.has_key('printformat'))&(data!=''):
                        #print '  key,printformat',key, descriptor['printformat']
                        line+=descriptor['printformat']%data
                    else:
                        line+=str(data)
                
                    line+=sep
                
                stringlist.append(line)
                
                
        # write to file
        file=open(file_name,'w')
        for line in stringlist:
            file.write(line+'\n')
            #print line
        file.close()
                      
                       
    
        
        
    def export_stops(self,file_name=''):
        """
        Exports data of current stops to
        spreadsheet formatted text file
        """
        sep='\t'

        
        if (not file_name)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            file_name=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()
            if not file_name: return

        elif not file_name:
            file_name=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfile_name))[0]+\
                                       '_stops'+'.txt'

        
        stringlist=['name'+sep+'zone'+sep+'id'+sep+'berth_no'+sep+'mode_exch'+sep+'coords']
        for stop in self.get_stops(all=1).values():
            stringlist.append(stop.name+sep+stop.zone+sep+stop.id+sep+\
                              str(stop.berthssec.berth_no)+sep+stop.mode_exch+sep+str(stop.get_pos_entrance()))
        
        
        # write to file
        file=open(file_name,'w')
        for line in self.get_exportfileheader(sep,name='Info on stops'):
            file.write(line+'\n')
            #print line
        
        for line in stringlist:
            file.write(line+'\n')
            #print line
        file.close()
            
            
    def export_results(self,file_name=''):
        """
        Exports results of current simulation to
        spreadsheet formatted text file
        """
        col_sep='\t'

        
        if (not file_name)&(self.mode=='gui'):
            if self.sim_flag:
                self.stop_gui()
            file_name=tkFileDialog.SaveAs(initialdir=self.simdir,\
                                          filetypes=[('Text files','*.txt'),\
                                                     ('All files','*')]).show()

        elif not file_name:
            file_name=self.simdir+os.sep+\
                 os.path.splitext(os.path.basename(self.simfile_name))[0]+\
                                       '_report'+'.txt'
            
        # initialize overall network data
        statistics_descriptor=[\
            {'key':'time',
             'name':'Simulated time',
             'type':'text',
             'width':1,
             'mode':'read',
             'valid':{'min':self.time,'max':10**8},
             'units':{'hh:mm:ss':1},
             },
            {'key':'length',
             'name':'Track length',
             'type':'real',
             'printformat':'%.2f',
             'width':1,
             'valid':{'min':0,'max':10**8},
             'units':{'m':1},
             'mode':'read',
             },
            {'name':'Number',
             'key':'number',
             'type':'integer',
             'printformat':'%9d',
             'width':1,
             'valid':{'max':10**8},
             'mode':'read'},
            {'name':'Number of vehicles',
             'key':'vehicles_total',
             'type':'integer',
             'printformat':'%9d',
             'width':1,
             'valid':{'max':10**8},
             'mode':'read'},
            {'name':'Total distance traveled',
             'key':'usertrippath_total',
             'type':'real',
             'printformat':'%9.1f',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.4,
             'units':{'km':1}
            },
            ]
        
        statistics_dataset={
            'time':self.seconds_to_string(self.time),
            'length':0.0,
            'number':0.0,
            'vehicles_total':0,
            'usertrippath_total':0.0}
        
       
        
        costs_descriptor=[{'name':'Estim. initial costs',
               'key':'estim_init',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
              {'name':'Estim. costs per year',
               'key':'estim_pa',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
              {'name':'Simulated Revenue',
               'key':'revenue',
               'type':'real',
               'printformat':'%9.2f',
               'width':1,
               'mode':'read',
               'valid':{'min':0,'max':10**8},
               'units':{self.currency:1},
               },
              ]
        
        costs_dataset={'estim_init':0.0,
                       'estim_pa':0.0,
                       'revenue':0.0}
              
        
        stringlist=self.get_exportfileheader(col_sep,name='Simulation results')
        
        
        # generate an ordered list for reports
        ids_ordered=self.objs.keys()
        ids_ordered.sort()
        
        for id in ids_ordered:
            obj=self.objs[id]
            #print '\n\n'+string.upper(obj.name)+'\n'
            # no net objects in list...not relevant, only confusing
            if self.split_id(obj.id)['category']!='net':
                stringlist.append('\n\n'+string.upper(obj.name))

                # do statistics
                descriptor,dataset=obj.report_statistics()
                if len(dataset)>0:
                    sumkeys=[]
                    sumdataset={}
                    for key in dataset.keys():
                        if key=='Total distance traveled':
                            if self.split_id(obj.id)['category']=='userman':
                                statistics_dataset['usertrippath_total']+=\
                                                               dataset[key]
                        if key in statistics_dataset.keys():
                            sumdataset[key]=dataset[key]
                            sumkeys.append(key)
                                    
                            # sum used for overall statistics 
                            if key=='number':
                                if self.split_id(obj.id)['category']=='car':
                                    #print '  count car',sumdataset[key]
                                    statistics_dataset['vehicles_total']+=\
                                                               sumdataset[key]
                                statistics_dataset[key]+=sumdataset[key]

                            
                            else:
                                statistics_dataset[key]+=sumdataset[key]
                    
                    # keys that are summed up are not shown in average values
                    for key in sumkeys:
                        del dataset[key]

                    if sumdataset.keys():
                        stringlist.append('\n'+'Total values')
                        self.dataset_to_sting(statistics_descriptor,sumdataset,
                                              stringlist,col_sep)
                    if dataset.keys():
                        stringlist.append('\n'+'Average values')
                        self.dataset_to_sting(descriptor,dataset,
                                              stringlist,col_sep)

                # do parameters
                descriptor,dataset=obj.report_params()
                if len(dataset)>0:
                    if dataset.keys():
                        stringlist.append('\n'+'Average parameters')
                        self.dataset_to_sting(descriptor,dataset,
                                              stringlist,col_sep)


                # do costs stuff
##                descriptor,dataset=obj.report_costs()
##                if len(dataset)>0:
##                    sumkeys=[]
##                    sumdataset={}
##                    for key in dataset.keys():
##                        if key in costs_dataset.keys():
##                            sumdataset[key]=dataset[key]
##                            sumkeys.append(key)
##                            # sum used for overall statistics 
##                            costs_dataset[key]+=sumdataset[key]
##
##                    # keys that are summed up are not shown in average values
##                    for key in sumkeys:
##                        del dataset[key]
##
##                    if sumdataset.keys():
##                        stringlist.append('\n'+'Costs')
##                        self.dataset_to_sting(costs_descriptor,sumdataset,
##                                              stringlist,col_sep)
##                    
##                    if dataset.keys():
##                        stringlist.append('\n'+'Average Costs')
##                        self.dataset_to_sting(descriptor,dataset,
##                                              stringlist,col_sep)

        stringlist.append('\n\nSUMMARY\n')

        # do not show total # of modules
        del statistics_dataset['number']
        stringlist.append('\nPerformance')
        self.dataset_to_sting(statistics_descriptor,statistics_dataset,
                              stringlist,col_sep)

##        stringlist.append('\nCosts')
##        self.dataset_to_sting(costs_descriptor,costs_dataset,
##                              stringlist,col_sep)
        # write to file
        file=open(file_name,'w')
        for line in stringlist:
            file.write(line+'\n')
            #print line
        file.close()
             
    #def format_results(self,obj,function):
    #    """
    #    Returns descriptor and dataset of formatted results of the obj,
    #    using the given formatting function.
    #    """
    #    pass
            
    def dataset_to_sting(self,descriptor,dataset,stringlist=[],separator='\t'):
        """
        Converts a dataset to a list of strings according to formatting
        information given in descriptor.
        """
        for descript in descriptor:
            key=descript['key']
            if dataset.has_key(key):
                if descript.has_key('name'):
                    name=descript['name']
                else:
                    name=key

                if descript.has_key('units'):
                    unit=separator+\
                          descript['units'].keys()[0] # take first available
                else:
                    unit=''

                if descript.has_key('printformat'):
                    value_string=descript['printformat']%(dataset[key])
                else:
                    value_string=repr(dataset[key])
                
                stringlist.append(name+separator+value_string+unit)

                
    def sim_step_gui(self,event):
        """
        Simulate a defined time step.
        """
        if not self.sim_flag:
            param=self.get_param()
            param['time_end']=self.time+0.5
            self.set_param(param)
            self.continue_gui()
            
    def set_param_gui(self):
        """
        Starts/refreshes panel to display and set simulation parameter.
        """
        if self.sim_flag:
            self.stop_gui()
            
        if not self.simpanel:
            self.simpanel=gui.Panel(None,
                                    title='Simulation Parameters',
                                    make_cmd=self.make_simpanel,
                                    refresh_cmd=self.refresh_simpanel,
                                    apply_cmd=self.apply_simpanel,
                                    close_cmd=self.close_simpanel)
        else:     
            self.simpanel.refresh()

    def make_simpanel(self,parent,widgets):
        """
        Create widgets for panel with simulation parameters.
        """
        descriptors=[
            {'key':'time',
             'name':'Current simulaton time',
             'type':'real',
             'printformat':'%8.2f',
             'width':1,
             'mode':'read',
             'valid':{'min':self.time,'max':10**8},
             'units':{'sec':1},
             },
            {'key':'time_end',
             'name':'End of simulation',
             'type':'real',
             'printformat':'%8.2f',
             'width':1,
             'mode':'write',
             'valid':{'min':0,'max':10**8},
             'units':{'sec':1},
             },
            # currently car precalculates some sample time dependent constants
            # during initialization. There must be a mechanism to recalculate
            # them upon changes of simulation parameters!
            {'key':'time_sample',
             'name':'Sample time',
             'type':'real',
             'printformat':'%1.4f',
             'width':1,
             'mode':'write',
             'valid':{'min':0.0001,'max':0.1},
             'units':{'sec':1},
             },
            {'key':'currency',
             'name':'Currency',
             'type':'text',
             'width':1,
                      'mode':'write',
             'valid':{'max':4}
             },
            {'key':'debug',
             'name':'Debug level',
             'type':'integer',
             'printformat':'%3d',
             'width':1,
             'mode':'write',
             'valid':{'min':0,'max':10},
             },
            ]

        widgets['parameters']=gui.DatasetEditor(descriptors,self.get_param(),
                                                parent)
        widgets['parameters'].pack(expand=1,fill='both')

    def refresh_simpanel(self,parent,widgets):
        """
        Refresh widgets for panel with simulation parameters.
        """
        pass

    def apply_simpanel(self,parent,widgets):
        """
        Apply widgets values for panel with simulation parameters.
        """
        param=self.get_param()
        newparam=widgets['parameters'].get_contents()

        # eliminate time, just to be sure that there will be no changes
        del newparam['time']

        
        # change time_sample for all objects (should go to set_param!!)
        if newparam['time_sample']!=param['time_sample']:
            for obj in self.objs.values():
                obj.change_time_sample(newparam['time_sample'])
        
        
        param.update(newparam)
        self.set_param(param)
        return 1

    def close_simpanel(self,parent,widgets):
        """
        Do something before simpanel is closed.
        Here the pointer to the panel is set to None.
        """
        
        self.simpanel=None
    
   
        
    def start_gui(self):
        start_sim=1
        #print 'start_gui',self.objs_iterate.keys(),self.objs_output.keys()
        if self.simpanel:
            if not self.sim_flag:
                start_sim=tkMessageBox.askyesno(
                    'Start Simulation?',
                    'Do you want to start the simulation\n'+
                    'with current simulation parameters')
                if start_sim:
                    self.simpanel.apply_and_close()
            
        if (not self.sim_flag)&start_sim:
            self.messagebar.message('state','Simulation running')

            # should recompile code...does not work like this
            # self.reload_obj()
            
            #Reset simulation time and all time-dependent statistics
            self.reset()
            
            # set the speed how often gui objects will be updated
            self.gui_update_count=self.gui_update_rate
            self.continue_flag=1

            # and go...
            self.loop_gui()

    def reset(self):
        """
        Reset simulation time and all time-dependent statistics
        """
        self.reset_time()
        for obj in self.objs.values():
            #print ' reset time and all statistical values of',obj.id
            obj.reset_statistics()

    def loop_gui(self):
        if self.maincycle():
            self.gui_update_count-=1
            if self.gui_update_count<0: # interrupt to update all widgets 
                self.simwin.after(0,self.update_gui)
            else:
                self.simwin.after_idle(self.loop_gui)
        else:
            
            comptime=time.time()-self.realtime_start
            simtime=self.time-self.time_start
            
            # do logging tasks here:
            if self.logtasks:
                self.close_log_data()
            
            self.messagebar.message('state','Simulation stopped at '+\
                                    self.seconds_to_string(self.time)+\
                                    ' simulated/computing time ratio='+\
                                    '%.2f'%(simtime/comptime))
            
            #print "Simulation completed in ",self.seconds_to_string(comptime)
            #print '  self.time',self.time
            #print '  self.time_start',self.time_start
            #print '  simtime',simtime
            #print '  ',
            #print '  time.time',time.time()
            #print '  self.realtime_start',self.realtime_start
            #print '  comptime',comptime
            #print '  ',
            #print '  ratio simtime/comptime',simtime/comptime
            #print '  ',
            #print '  self.time_comp',self.time_comp
            #print '  self.time_end',self.time_end

            
    def update_gui(self):
        """ update all widgets"""
        self.messagebar.message('systemevent','Simulation running for '+\
                                self.seconds_to_string(self.time))
        
        # do logging tasks here:
        if self.logtasks:
            self.log_data()
            
        
        self.simwin.update_idletasks()
        self.gui_update_count=self.gui_update_rate
        self.simwin.after_idle(self.loop_gui)

    def get_duration(self,unit='s'):
        """
        Return duration of current simulation as a real number
        in the requested unit.
        """
        if unit=='s':
            return self.time
        elif unit=='h':
            return self.time/3600
        elif unit=='min':
            return self.time/60
        else: return self.time
            
        
    def seconds_to_string(self,seconds):
        minutes,secs=divmod(seconds,60)
        hours,mins=divmod(minutes,60)
        return '%d:%02d:%02d'%(hours,mins,secs)

    def start(self,time_end=-1):
        if time_end>=0:
            # this is messy because set_param in messy
            self.time_end=time_end
            self._param['time_end']=self.time_end
            
        if self.objs:
            # continue the simulation from the time when
            # the simulation file has been saved
            self.prepare_continue()
            if self.debug>=3:
                print 'Continue simulation at time [hh:mm:ss]',\
                      self.seconds_to_string(self.time_start),'.'
                
            while self.maincycle():
                # do logging tasks here:
                
                if self.logtasks:
                    self.log_data()

            if self.debug>=3:
                comptime=time.time()-self.realtime_start
                simtime=self.time-self.time_start
                
                print 'Completed simulation at time [hh:mm:ss]',\
                      self.seconds_to_string(self.time),'.'
                print 'Computing time has been '+\
                      '%.2f'%(simtime/comptime)+' times faster than real time.'
        else:
            if self.debug>=3:
                print 'No simulation objects present.'
    
    def loop_cml(self):
        if self.simfile_name:
            # jump into simulation loop
            self.start()

            # write report
            if self.filename_report:
                print 'save report to file:',self.filename_report
                self.export_results(self.filename_report)
                
            # save results
            if self.filename_output:
                print 'save simulation to outputfile:',self.filename_output
                #try:
                self.save(self.filename_output)
                #except:
                #    print 'Could not save to outputfile',self.filename_output
            
            # do logging tasks here:
            if self.logtasks:
                self.close_log_data()
                
        else:
            print "No simulation filename defined"
            sys.exit()
       
       
    def init_sim(self,param_cml):
        '''
        Reads self.simfile (if exists) and gets ojects initialized
        The param_cml include the command line parameters, but contain also
        other parameters added by __main__ or Mainwin
        '''
        
        # include these simulation objects if no simfile is defined
        self.sim_data_default=[{'version':self.__version__},
                               {'type':'base.net.netedit_gui'},
                               {'type':'base.net.netmon'},
                               {'type':'base.net.netsim'},
                               ]
        
        if (self.simfile_name!='untitled.its')&(self.simfile_name!=''):
            
            simfile=open(self.simfile_name,'r')
            # print 'pickle.load...'
            sim_data=pickle.load(simfile)
            simfile.close()
            if len(sim_data)==0:
                print 'WARNING: No data in file ',simfile_name
                sim_data=self.sim_data_default    
                
                        
        else:
            sim_data=self.sim_data_default
            
        # start object initialization
        self.init_flag=1

        # parameter setting...
        
        # overwrite global sim parameters with parameters of simfile
        # which is the first item of the  sim_data list
        param=(sim_data[0])

        # finally overwrite these parameters with command line parameters
        param.update(param_cml)

        # ...and set them
        self.set_param(param)

        
        # read states of each simulation object, initialize and append
        for state_image in sim_data[1:]:
            if debug>=5:
                print 'init_sim: state_image for',state_image['type'],\
                      ', version',state_image['version']
                for key in state_image.keys():print '  ',key 

            # append, initialize for sim but don't link
            if state_image.has_key('mods'):
                # only append if simobject with modules contains modules
                if state_image['mods']:
                    self.append_obj(state_image['type'],
                                    state_image,link_flag=0)
            elif state_image['type']=='base.net.odgen':
              # never this one during initialization
              pass
            
            else:
                # try to append all other modules
                self.append_obj(state_image['type'],
                                    state_image,link_flag=0)

        self.link_objs() # link now, alltogether
        self.init_flag=0
                
    def append_obj(self,type,state_image=None,link_flag=1):
        '''
        Initialize Simulation object and add it to
        simulation objects table.
        Link also if link_flag is set
        '''

        # check whether obj is not already appended
        if not self.objs.has_key(type):
            Obj,libs=self.import_obj(type)

            if not Obj: # import failed
                return None
          
            obj=Obj(type,state_image,self)
            
            
            # gui objects will be included and initialized, but not linked!!
            self.objs[type]=obj
            self.libs[type]=libs
            if self.debug>3:
                
                # after set_state, the version of the object itself
                # may have been upgraded already
                if state_image!=None:
                    version=state_image.get('version',self.__version__)
                else:
                    version=self.__version__
                    
                # if there is no state image, the version is the one of sim
                print "Appended "+type+', version',version
                          

            if link_flag:
                self.link_objs()

            return obj # return newly initialized object
        
        else: # object already initialized
            return self.objs[type]
    
    def append_obj_iterate(self,obj):
        #print 'append_obj_iterate:'
        # check if obj is a current simulation object
        if obj:#self.objs.has_key(obj.type):
            #print '  append_obj_iterate: try with',obj.type
            # check first if obj is not already among the iterate objects
            if not self.objs_iterate.has_key(obj.type):
                self.objs_iterate[obj.type]=obj
                self.iterate_count=int(obj.iterate_interval/self.time_sample)
                #print '  do append_obj_iterate',obj.type
        else:
            # object is probably a None object that came up with some
            # preview initializations
            pass
            #print 'WARNING: append_obj_iterate: illigal append of',obj
        
    def append_obj_output(self,obj):
        # check if obj is a current simulation object
        if obj:#self.objs.has_key(obj.type):
            # check first if obj is not already among the output-objects
            if not self.objs_output.has_key(obj.type):
                self.objs_output[obj.type]=obj
                obj.output_count=int(obj.output_interval/self.time_sample)
        else:
            # object is probably a None object that came up with some
            # preview initializations
            pass
            #print 'WARNING: append_obj_output: illigal append of',obj.id

    def remove_obj(self,obj):
        """
        Remove object from current simulation object list.
        """
        
        if self.objs.has_key(obj.type):
            self.remove_obj_iterate(obj)
            self.remove_obj_output(obj)
            if self.debug>=5:
                print 'Remove simulation object '+obj.type
            del self.objs[obj.type]
            del self.libs[obj.type]
        
            
    def remove_obj_iterate(self,obj):
        # check if obj is a current simulation object
        if self.objs.has_key(obj.type):
            # check first if obj is among the iterate objects
            if self.objs_iterate.has_key(obj.type):
                del self.objs_iterate[obj.type]
        
        
    def remove_obj_output(self,obj):
        # check if obj is a current simulation object
        if self.objs.has_key(obj.type):
            # check first if obj is among the output-objects
            if self.objs_output.has_key(obj.type):
                del self.objs_output[obj.type]
            
    def link_objs(self):
        for obj in self.objs.values():
            # include obj, except for gui object in non-gui mode
            if not((obj.gui_flag==1) & (self.mode!="gui")):
                #print 'link_obj',obj.id
                obj.link()
        for obj in self.objs.values():
            # include obj, except for gui object in non-gui mode
            if not((obj.gui_flag==1) & (self.mode!="gui")):
                #print 'rebuild_pointers',obj.id
                obj.rebuild_pointers()

                # upgrade version here
                obj.__version__=self.__version__

    def activate_all(self):
        """
        Activate all track elemets.
        """
        for id in self.objs.keys():
            id_data=self.split_id(id)
            if (id_data['category']=='track'):
                for elem in self.objs[id].mods.values():
                    elem.set_status('active')
                

    def inactivate_all(self):
        """
        InActivate all track elemets.
        """
        for id in self.objs.keys():
            id_data=self.split_id(id)
            if (id_data['category']=='track'):
                for elem in self.objs[id].mods.values():
                    elem.set_status('passive')
    

    def reload_obj(self,obj=None):
        '''
        Currently, this has NO effect on the simulation!!!???
        
        Reload all or a particular object in memory.
        In this way edited skripts can get active without restarting
        the simulator.
        '''
        if not obj:
            for libs in self.libs.values():
                for lib in libs:
                    reload(lib)
        else:
             for lib in self.libs[obj.type]:
                 reload(lib)
                
    def maincycle(self):
        # This is the cycle that triggers all time-driven events
        # of all  simulation objects that are in the objs_iterate
        # or objs_output table
        self.maincycle_flag=1

        # call iterate methods...if necessary
        for obj in self.objs_iterate.values():
            if obj.iterate_interval==0.0:
                obj.iterate()
            else:
                obj.iterate_count-=1
                if obj.iterate_count<0:
                    obj.iterate_count=int(
                        obj.iterate_interval/self.time_sample)
                    obj.iterate()
                    
        # call output methods...if necessary
        for obj in self.objs_output.values():
            if obj.output_interval==0.0:
                obj.output()
            else:
                obj.output_count-=1
                if obj.output_count<0:
                    obj.output_count=int(obj.output_interval/self.time_sample)
                    obj.output()
        self.maincycle_flag=0

        # increases time and returns 1 if simulation has not terminated
        self.time += self.time_sample
        
        self.sim_flag=(self.time < self.time_comp) and  self.continue_flag
        # print self.time,self.time_comp,self.continue_flag,self.sim_flag

        # returns 0 when end of simulation is reached
        return self.sim_flag


    def get_type(self,family,category,obj_name):
        """
        Return type.
        """
        return family+type_sep+category+type_sep+obj_name

    
    def get_id(self,type='',mod_no=-1):
        """
        Return id
        """
        return type+id_sep+`mod_no`

    def get_mod(self,id):
        """
        Get a module with the specified id. 
        """
        return self.split_id(id)['mod']

    def get_ids(self,objs):
        """
        Returns a list of the ids of the objects or modules
        in the objs list.
        """
        ids=[]
        for obj in objs:
            if type(obj)!=types.StringType:
                ids.append(obj.id)
            else:
                # obj seems to be already an id
                ids.append(obj)

        return ids
                
                

    def rebuild_pointer(self,id):
        """
        Try to rebuild pointer from id, but only if valid.
        """
        if type(id)==types.StringType:
            mod=self.get_mod(id)
            if mod!=None:
                return mod
            else:
                return id
        else:
            return id
        
    def rebuild_pointers(self,ids):
        """
        Try to rebuild a list of pointers id, but only if valid.
        """
        pointers=[]
        for id in ids:
            pointers.append(self.rebuild_pointer(id))

        return pointers
        
    def split_id(self,tag):
        """
        helper function to decompose a tag string that contains family
        name, type, id, e.t.c. All elements will be stored and returned in
        a table.
        """

        family=None
        category=None
        obj=None
        serial_data=None
        mod=None
        name=''
        part_id=None
        part_type=None
        part_no=None
        obj_type=''
        error=0
        
        # pars id and extract information
        
        tag_elements=string.split(tag,id_sep)
        #print 'split_id tag_elements',tag_elements
        if len(tag_elements)==2:
            obj_type, serial_data=tag_elements
            type_elements=string.split(obj_type,type_sep)
            family=type_elements[0]
            category=type_elements[1]
            name=type_elements[2]+' '+tag_elements[1]
            # print 'split_id type_elements',type_elements
            
            if self.objs.has_key(obj_type):
                obj=self.objs[obj_type]
            
            serial_elements=string.split(serial_data,part_sep)
            if len(serial_elements)==3:
                mod_no_str,part_type,part_no_str=serial_elements
                part_id=part_type+part_no_str
                part_no=string.atoi(part_no_str)
            
            elif len(serial_elements)==2:
                mod_no_str,part_id=serial_elements
                
            elif len(serial_elements)==1:
                mod_no_str=str(serial_elements[0])
                
            mod_no=string.atoi(mod_no_str)
            if obj:
                if obj.multimod_flag:
                    if obj.mods.has_key(mod_no):
                        mod=obj.mods[mod_no]
                    
        elif len(tag_elements)==1:
            obj_type=tag_elements[0]
            type_elements=string.split(obj_type,type_sep)
            if len(type_elements)>2:
                family=type_elements[0]
                category=type_elements[1]
                name=type_elements[2]
            else:
                error=1
                #print 'ERROR in Sim.split_id: What kind of tag is this:',tag
                #print 'type_elements=',type_elements

            if self.objs.has_key(obj_type):
                obj=self.objs[obj_type]
                
        tag_split={'family':family,'obj':obj,'mod':mod,'category':category,
                   'part_id':part_id,'part_type':part_type,'part_no':part_no,
                   'serial_data':serial_data,'type':obj_type,'name':name,
                   'error':error}

        return tag_split


    def import_obj(self,type):
        libs=[]
        comps=string.split(type,type_sep)
        #try:
        lib=__import__('objlib.'+string.join(comps[:],'.'))
        libs.append(lib)
        for comp in comps:
            lib= getattr(lib,comp)
            #print lib
            libs.append(lib)
        # return class of simulation object
        Obj=getattr(lib,string.capitalize(comps[-1]))

        #except:
        #    print 'WARNING in Sim.import_obj: initialiasation of',\
        #          type,'failed'
        #    Obj=None
        #    libs=None
        return Obj, libs

    
    def reset_time(self):
        self.time_start=0.0
        self.time=0.0
        self.time_comp=self.time_end
        self.realtime_start=time.time() # save start time in seconds
        self.sim_flag=self.time < self.time_comp
        
    
    #def get_dir_name_ext(self,name):
    #    """
    #    Returns the directory, the filename without extension and
    #    the extension.
    #
    #    Example: 'its/projects/fiera.its' => 'its/projects','fiera','its'
    #    """
    
    def save(self,simfile_name):
        """
        Saves state of current simulation, ioncluding all objects.
        Functions and guo objects are not saved and must be reconstructed.
        """
        self.set_simfile_name(simfile_name)

        # save all global simulation parameters 
        simdata=[self.get_param()]
        #print 'save',simdata
        try:
            simfile=open(simfile_name,'w')
        except:
            print 'WARNING in save: could not open',simfile_name
            return 1
        
        # get state_image of all objects and append to list
        for obj in self.objs.values():
            if obj.id !='base.net.netsim':
                state_image=obj.get_state_image()
                simdata.append(state_image)

        state_image=self.objs['base.net.netsim'].get_state_image()
        simdata.append(state_image)
    
        # dump serialized data to file
        #try:
        # bin=1 causes errors in windows!!
        pickle.dump(simdata, simfile,bin=0) 
        #except:
        #    print 'WARNING in save: could not dump data in',simfile_name
        #    #return 1
        
             
        simfile.close()

        # now open a zip file and zip it
        #zipfile_name=os.path.dirname(simfile_name)+os.sep+\
        #         os.path.splitext(os.path.basename(simfile_name))[0]+'.zip'
        #print 'zip:', zipfile_name
        #file_zip=zipfile.ZipFile(zipfile_name,'w')
        #file_zip.write(simfile_name,simfile_name+'zipped')
        #file_zip.close()
        
        

        
        # reinitialize objects in order to set again pointers to functions
        for state_image in simdata[1:]:
            obj=self.objs[state_image['type']]
            obj.set_state_image(state_image)

        self.init_flag=1  
        self.link_objs() # link now, alltogether
        self.init_flag=0

        return 0
        
    def configure_objectlibbrowser(self,**kw):
        if not self.objectlibbrowser:
            self.objectlibbrowser=gui.Objectlibbrowser_simple(\
                self,callback_on_close=self.close_objectlibbrowser)
        
        self.objectlibbrowser.configure(**kw)
            
    def close_objectlibbrowser(self):
        #print 'close_objectlibbrowser'
        self.objectlibbrowser=None
            
    

class Mainwin:
    """
    Main root window object of the its gui application.
    Mainwin can initialize and run differen applications.
    However, at present the simulator (Sim class) is the only one.
    """
    def __init__(self, parent, simfile_name,**args):

        # determine parameters dependent if they where passed by
        # key-value pairs or by a single parameter dictionary
        if args.has_key('params'):
            self.param=args['params']
        else:
            self.param=args

        # the initial simulation handle
        self.sim=None
        
        
        
        self.__version__=__version__ #param['version']

        # fileopendialog widget:
        #self.fileopendialog=None
        
        # directories and file styoes
        self.simdir=os.getcwd()+os.sep+'projects'

        #self.file_type='*.its'
        self.file_types=[('Sim files','*.its'),('All files','*')]
        # Create the toplevel to contain the main menubar.
        #self.megatoplevel=Tkinter.Toplevel()
        #toplevel = self.megatoplevel
        self.megatoplevel = Pmw.MegaToplevel(parent)
        self.megatoplevel.protocol('WM_DELETE_WINDOW',self.exit)
        
	self.toplevel = self.megatoplevel.interior()
        
	# Create the Balloon for this toplevel.
	self.balloon = Pmw.Balloon(self.toplevel)

	# Create and install the MenuBar.
	menubar = Pmw.MainMenuBar(self.toplevel,
                balloon = self.balloon)
        self.toplevel.configure(menu = menubar)
	self.menubar = menubar

	# Add basic menues.
	menubar.addmenu('File', 'Operations on current simulation file')

	menubar.addmenuitem('File', 'command', 'Create a new simulation file',
		command = self.new_file,
		label = 'New')

        
        menubar.addmenuitem('File', 'command', 'Open new simulation file...',
                            command = self.open_file,
                            label = 'Open...')
        
        
	menubar.addmenuitem('File', 'separator')

        menubar.addmenuitem('File', 'command', 'Safe simulation file as ...',
                            command = self.saveas_file,
                            label = 'Save as...')

        menubar.addmenuitem('File', 'command',
                            'Safe simulation under current name',
                            command = self.save_file,
                            label = 'Save')
        
        menubar.addmenuitem('File', 'separator')

        #menubar.addmenuitem('File', 'command', 'Close current simulation',
	#	command = self.close_file,
	#	label = 'Close')

	menubar.addmenuitem('File', 'command', 'Exit the application',
		command = self.exit,
		label = 'Exit')

        # does not work:
	#Pmw.setbusycursorattributes(self.megatoplevel.component('hull'),
        #                            exclude = 1)

	# add help menue
	#menubar.addmenu('Help', 'User manuals', name = 'help')
	#menubar.addmenuitem('Help', 'command', 'About this application',
	#	command = PrintOne('Action: about'),
	#	label = 'About...')

        

	# Create and pack the main part of the window.
	self.simwin =Tkinter.Frame(self.toplevel,bd=2,relief='sunken')
        #self.simwin.pack(expand=1, fill='both',side='right')
        empty=Tkinter.Canvas(self.simwin,width = 1.6*300,
                                         height = 300)
        empty.pack(fill = 'both', expand = 1)
        
        self.simwin.pack(expand=1, fill='both')
                
	# Create and pack the MessageBar.
	self.messagebar = Pmw.MessageBar(self.toplevel,
		entry_width = 40,
		entry_relief='groove',
		labelpos = 'w',
	        label_text = 'Status:')
        
	#self.messagebar.pack(fill = 'x', padx = 10, pady = 10)
        self.messagebar.pack(fill = 'x')
	self.messagebar.message('state','')
        self.messagebar.component('entry').configure(relief='sunken')
        

	
	# Configure the balloon to displays its status messages in the
	# message bar.
	self.balloon.configure(statuscommand = self.messagebar.helpmessage)

	self.testMenuList = []
        if simfile_name:
            # initialize simulator
            self.start_sim(simfile_name)
        else:
            self.new_file()

        empty.destroy()
        
    def start_sim(self,simfile_name):
        """
        Initialize the simulation object.
        """
        
        self.messagebar.message('busy','Starting'+
                                        repr(simfile_name))
        self.megatoplevel.update()
        self.simfile_name=simfile_name
        # set window title
        self.set_wintitle(simfile_name)

        if os.path.dirname(simfile_name):
            self.simdir=os.path.dirname(os.path.abspath(simfile_name))
            
        # initialize Simulator
        self.sim=Sim(simfile_name,
                     params=self.param,
                     simdir=self.simdir,
                     simwin=self.simwin,
                     menubar=self.menubar,
                     balloon=self.balloon,
                     messagebar=self.messagebar)

        self.messagebar.resetmessages('busy')
        self.messagebar.message('systemevent','Done')
        
    def set_wintitle(self,simfile_name=''):
        '''Set window title'''
        if simfile_name!='':
            titlefile_name=os.path.basename(simfile_name)
        else:
            titlefile_name=''

        self.megatoplevel.title(titlefile_name+' - its'+self.__version__)

    def save_current_file(self):
        """
        If there is a file open, it asks to save it.
        If the cancel button has been pressed, it will return a zero,
        otherwise one.
        """

        if self.sim:
            if self.sim: #.save_flag:
                save_it=tkMessageBox.askyesno(
                    'Save file?',
                    'Do you want to save current file?')
                
                #save_it=gui.ask('There is file open.\n Do you want to save it',
                #                bottons=('Save', 'Save As...','No','Cancel'))
                #print 'Do you want to save current file?',save_it
                if save_it:
                    if self.simfile_name=='untitled.its':
                        return self.saveas_file()
                    else:
                        return self.save_file()
                    
                else: return 1

            else: return 0
                
        else: return 1
        
            
           
            
    def new_file(self):
        save_it='Ok'
        if self.save_current_file():
            self.close_file()
            # initialize Simulator
            self.start_sim('untitled.its')
         
    def open_file(self):
        
        if self.save_current_file():
            #self.fileopendialog=tkFileDialog.Open()
            
            #ask_openfilename=gui.PmwFileDialog()
            #ask_openfilename.title('Open File')
            simfile_name =tkFileDialog.askopenfilename(
                initialdir=self.simdir, filetypes=self.file_types)


            #tkFileDialog.ask_openfilename.askfilename(filter=self.file_type,
            #                                            directory=self.simdir,
            #                                            )
        
            if simfile_name:
                self.close_file()
                # initialize Simulator
                self.start_sim(simfile_name)
            
            
    def saveas_file(self):
        """
        Ask for a filename and save current simulation.
        Returns one if everything went successful and zero if canceled.
        """
        simfile_name=tkFileDialog.SaveAs(initialdir=self.simdir,
                                         filetypes=self.file_types).show()

        # be sure that .its extension has been added
        s=string.split(simfile_name,'.')
        if s[-1]=='ITS':
            s[-1]='its'
        elif s[-1]!='its':
            s.append('its')

        simfile_name=string.join(s,'.')
            
        
        #ask_filename=gui.PmwFileDialog()
        #ask_filename.title('Save File As')
        #simfile_name = ask_filename.askfilename(filter=self.file_type,
        #                                        directory=self.simdir)
        
        if simfile_name:
            #self.simdir=os.path.dirname(simfile_name)
            #save_it='Yes'
            #if os.path.isfile(simfile_name):
            #    #print 'File exists !!!!!!!'
            #    save_it=gui.ask('Overwrite exiting file')
                
            #if save_it=='Yes':
                # save state of simulation
                #Pmw.showbusycursor()
            self.messagebar.message('busy','Saving'+
                                    repr(simfile_name))
            self.megatoplevel.update()
            self.sim.save(simfile_name)
            self.megatoplevel.update()
            #Pmw.hidebusycursor()
            self.set_wintitle(simfile_name)
            self.messagebar.resetmessages('busy')
            self.messagebar.message('systemevent','Done')
            return 1
        else:
            return 0
        

    def save_file(self):
        if self.sim:
            self.messagebar.message('busy','Saving'+
                                        repr(self.sim.simfile_name))
            
            self.megatoplevel.update()
            # save state of simulation
            self.sim.save(self.sim.simfile_name)

            self.megatoplevel.update()
            
            self.messagebar.resetmessages('busy')
            self.messagebar.message('systemevent','Done')
            return 1
        else:
            return 0
        
    def close_file(self):
        if self.sim:
            #if self.save_current_file():
            self.sim.close()
            del self.sim

            # define None simulation handle
            self.sim=None

            # set window title
            self.set_wintitle('No application loaded')
            self.messagebar.message('state','No application loaded')

    def exit(self):
        if self.sim:
            if self.save_current_file():
                self.sim.close()
                sys.exit()
                
    #------------------------------------ yunk
    def add(self):
	if len(self.testMenuList) == 0:
	    num = 0
	else:
	    num = self.testMenuList[-1]
	num = num + 1
	name = 'Menu%d' % num
	self.testMenuList.append(num)

	self.menubar.addmenu(name, 'This is ' + name)

    def delete(self):
	if len(self.testMenuList) == 0:
	    self.menubar.bell()
	else:
	    num = self.testMenuList[0]
	    name = 'Menu%d' % num
	    del self.testMenuList[0]
	    self.menubar.deletemenu(name)

    def additem(self):
	if len(self.testMenuList) == 0:
	    self.menubar.bell()
	else:
	    num = self.testMenuList[-1]
	    menuName = 'Menu%d' % num
            menu = self.menubar.component(menuName)
            if menu.index('end') is None:
                label = 'item X'
            else:
                label = menu.entrycget('end', 'label') + 'X'
            self.menubar.addmenuitem(menuName, 'command', 'Help for ' + label,
                    command = PrintOne('Action: ' + menuName + ': ' + label),
                    label = label)
            
    def deleteitem(self):
	if len(self.testMenuList) == 0:
	    self.menubar.bell()
	else:
	    num = self.testMenuList[-1]
	    menuName = 'Menu%d' % num
            menu = self.menubar.component(menuName)
            if menu.index('end') is None:
                self.menubar.bell()
            else:
                self.menubar.deletemenuitems(menuName, 0)
            



###############################################################################
if __name__ == '__main__':
    """
    This is the launcher program for the simulation.
    It will essentially start the simulation in graphical mode,
    which is the default, or command-line mode.
    It dous also read and set all other commandline options.
    """

    
    
    # db for simulation parameters
    params={'mode':'gui',
           }
    
    try:
        # scan command line arguments
        optlist, args = getopt.getopt(
            sys.argv[1:], 'hct:d:e:s:o:fr:',
            ['help','h','cml','output=','feedback','report='])
    except getopt.GetoptError:
        print usage
        sys.exit(2)
        
    if len(args)==1:
        simfile_name=args[0]
    elif len(args)==0:
        simfile_name=''
    else:
        print usage
        sys.exit()

    for opt in optlist:
        if (opt[0]=='-h')|(opt[0]=='--h')|(opt[0]=='--help'):
            print usage
            sys.exit()

        elif (opt[0]=='-c')|(opt[0]=='--cml'):
            # command line mode
            params['mode']='cml'

        #elif (opt[0]=='-a')|(opt[0]=='--alpha'):
        #    # alpha terminal mode
        #    params['mode']='alpha'

        elif (opt[0]=='-e'):
            # simulate until the given time
            params['time_end']=string.atof(opt[1])

        elif (opt[0]=='-d'):
            # simulate for the given time in seconds
            params['time_delta']=string.atof(opt[1])
            
        elif (opt[0]=='-s'):
            params['time_sample']=string.atof(opt[1])

        elif (opt[0]=='-o')|(opt[0]=='--output'):
            # save simulation to this file after exiting
            params['filename_output']=opt[1]
           
        elif (opt[0]=='-f')|(opt[0]=='--feedback'):
            # results are fed back to input simfile 
            params['filename_output']=simfile_name

        elif (opt[0]=='-r')|(opt[0]=='--report'):
            # save simulation to this file after exiting
            params['filename_report']=opt[1]


    
    
    # launch the software dependent on mode
    if params['mode']=='gui':
        rootwin = Tkinter.Tk()
        rootwin.withdraw()
        Pmw.initialise(rootwin, fontScheme = 'pmw1')
        #Pmw.initialise(rootwin, fontScheme = 'pmw2')
       
        Mainwin(rootwin,simfile_name,params=params)
        #print 'Pmw',Pmw.__dict__.keys()
        rootwin.mainloop()

    else:
        Sim(simfile_name,params=params)
 
