#Copy file in same directory as its.py and run it with
#python its3d.py projects/simfile.its

usage="""
USAGE:
    Copy this file in same directory as its.py and run it with
    python its3d.py projects/simfile.its

    If no simfile is given, the default file defined by the
    variable simfile_name is used, you can modify this variable in the
    script file.

    Command line options are:
    -h, --help showth this help message
    -p         name of type of pole to be used for elevated guideway
               valid names are: round, arc, portici
               Defaults can also modified in the scriptfile.

    Usage of 3D Monitor
    Initially you will see the network from above in observer mode.
    You have the following function-keys:

    Key      Function
    r        switch to observer external observer mode.
    u        sit on first vehicle
    b        move with first vehicle but below the guideway
    s        stop simulation
    c        continue simulation
    Esc      Quit application

    The following navigation keys are only valid for observer mode.
    8        move forward
    2        move backward
    6        move right
    4        move left
    9        move in
    3        move out
    To rotate around z and y axes: double-klick in 3D window and move mouse
    while keeping left mouse-button pressed.
    """
###############################################################################
# DEFINE DEFAULTS HERE

#simfile_name='projects/fiera_users.its'
simfile_name='projects/rings2_cars60_users30.its'

polename='round'
#polename='arc'
#polename='portici'

###############################################################################
# IMPORT
import its
import sys
from OpenGL.GL import *
from OpenGL.GLE import *
from OpenGL.GLUT import *
from Numeric import *
from RandomArray import *
from LinearAlgebra import determinant
import getopt


def loop_gl(arg):
    if sim.sim_flag:
        sim.maincycle()
    
###############################################################################
if __name__ == '__main__':
    """
    This is the launcher program for OpenGL graphical visualization
    """
    params={}
    
    try:
        # scan command line arguments
        optlist, args = getopt.getopt(
            sys.argv[1:], 'hctp:d:e:s:o:fr:',
            ['help','h','cml','output=','feedback','report='])
    except getopt.GetoptError:
        print usage
        sys.exit(2)
        
    if len(args)==1:
        simfile_name=args[0]
    elif len(args)==0:
        pass
    else:
        print usage
        sys.exit()

    for opt in optlist:
        if (opt[0]=='-h')|(opt[0]=='--h')|(opt[0]=='--help'):
            print usage
            sys.exit()

        elif (opt[0]=='-p'):
            #print '  polename',opt[1]
            params['polename']=opt[1]
     
        elif (opt[0]=='-e'):
            # simulate until the given time
            params['time_end']=string.atof(opt[1])

        elif (opt[0]=='-d'):
            # simulate for the given time in seconds
            params['time_delta']=string.atof(opt[1])
            
        elif (opt[0]=='-s'):
            params['time_sample']=string.atof(opt[1])

        elif (opt[0]=='-o')|(opt[0]=='--output'):
            # save simulation to this file after exiting
            params['filename_output']=opt[1]
           
        elif (opt[0]=='-f')|(opt[0]=='--feedback'):
            # results are fed back to input simfile 
            params['filename_output']=simfile_name

        elif (opt[0]=='-r')|(opt[0]=='--report'):
            # save simulation to this file after exiting
            params['filename_report']=opt[1]


    
    
    # initialize GLUT 
    glutInit(sys.argv)
    
    # initialize the simulator with the simulation file
    sim=its.Sim(simfile_name,params=params)

    print 'Append OpenGL(R) 3D monitor object'
    netmon_gl=sim.append_obj('base.net.netmon_gl',link_flag=0)

    if params.has_key('polename'):
        #print 'set default valye from command line'
        netmon_gl.set_poletype(params['polename'])
    else:
        #set default
        netmon_gl.set_poletype(polename)

    # show users or not
    netmon_gl.show_users_flag=0
    
    # link now with other ojects...and generate matrices
    netmon_gl.link()

    
    sim.prepare_continue()

    
    
    # enter the event loop
    glutMainLoop ()
