###############################################################################
# IMPORTS
from classlib.base import * #Obj,Array_db,Obj_db
import copy
###############################################################################



class Node:
    """
    Contains all information of a specific (open) node. The methods are common
    to all nodes and linked to functions that are implemented in Netsim.
    """
    def __init__(self,netsim,localnode,sec):
        
        self.localnode=localnode
        self.sec=sec

        # absolute position (x,y,z) in array format:
        self.pos=localnode['trans'][0:3,3]
        
        # will be determined by add_node
        self.id=0
        
        netsim.add_node(self)
        self.netsim=netsim

        
class Netsim(Obj):
    def init_obj(self):
        if not self.sim.objs.has_key(self.type):
            self.output_interval=0.2
                
            self.secs={} 
            self.sec_id_current=0

            self.segs=Array_db()
            self.segs_total=Array_db()

            self.nodes_open={}
            self.node_current = 0

            self.cars=Obj_db()
            self.path=Array_db()
            self.path_enter=Array_db()
            self.path_sec=None
            self.length=Array_db()
            self.velocity=Array_db()
            self.segs_ind=Array_db()
            self.seg_length=Array_db()

            #self.v=Array_db()
            # This is an array with brake factors which must result in the
            # braking distance when multiplied with the squared velocity.
            # For each carrier there is a row, containing the following values:
            # [ 1/(2*a_comfort),1/(2*a_emergency),1/(2*a_failure)]
            # where the three constance multiplied with the squared velocity
            # give the comfort braking distance, the emergency braking distance
            # and the the failure braking distance, respectively.
            self.brake_factors=Array_db()
            self.times_brake=Array_db()
            self.brakedists=array([])

    #def rebuild_secpointers(self,secs):
    #    """
    #    This function is called from the elem.rebuild_pointers method,
    #    in order to rebuild netsim.
    #    secs is the section list of the respective track element.
    #    """
    #    for sec in secs:
    #        if self.secs.has_key(sec.id):
    #            self.secs[sec.id]=sec
    #            

    def get_state(self):

        
        # clean nodes from objects
        for node in  self.nodes_open.values():
            #node.sec=sec.id
            del node.netsim


        # this is somewhat curious:
        # the section pointers are replaced by the id of their element
        # this is sufficient information for rebuild_pointers
        for secid in self.secs.keys():
            if type(self.secs[secid])!=types.StringType:
                self.secs[secid]=self.secs[secid].elem.id
            else:
                print 'WARNING: get_state in',self.id
                print '  found',self.secs[secid],' instead of sec object',secid

        # replace car objects by their id
        self.cars.objs_to_ids(self.sim)
        
        return {'version':self.__version__,
                'secs':self.secs,
                'sec_id_current':self.sec_id_current,
                'segs':self.segs,
                'segs_total':self.segs_total,
                'nodes_open':self.nodes_open,
                'node_current':self.node_current,
                'cars':self.cars,
                #'carobjs':self.carobjs,
                #'car_ids':self.car_ids,
                'path':self.path,
                'path_enter':self.path_enter,
                'path_sec':self.path_sec,
                'length':self.length,
                'velocity':self.velocity,
                'segs_ind':self.segs_ind,
                'seg_length':self.seg_length,
                'brake_factors':self.brake_factors,
                'brakedists':self.brakedists,
                'times_brake':self.times_brake}

    def rebuild_pointers(self):
        #print 'netsim.rebuild_pointers...'
        # make sections first
        for secid in self.secs.keys():
            self.secs[secid]=self.rebuild_secpointer(secid)
            #print '  rebuilt', self.secs[secid].id,self.secs[secid].elem.id
            #print '    tosecs_active=',self.secs[secid].tosecs_active
            #if len(self.secs[secid].tosecs_active[0])>0:
            #    print '     seg[0]=',self.secs[secid].tosecs_active[0].segs[0]
            #else:
            #    print '     no actve secs'
            
        # rebuild  node pointers
        for node in  self.nodes_open.values():
            # check if node.sec has been replaced by id string 
            if type(node.sec)==types.StringType:
                # check if section has been already replaced by object
                node.sec=self.secs[node.sec]
            node.netsim=self

        if 1:
            #print 'rebuild_pointers',self.id
            
            for sec in self.secs.values():
                #print '  correct segment angles here.',sec.id
                if (sec.status=='active')&(len(sec.tosecs_active)>0):
                    #print '  correct sec',sec.id,sec.elem.id
                    for n_seg in range(0,len(sec.segs)-1):
                        trans0=sec.segs[n_seg]
                        if n_seg==len(sec.segs)-2:
                            trans1=sec.tosecs_active[0].segs[0]
                        else:
                            trans1=sec.segs[n_seg+1]
                        phi=arctan2(\
                        (sec.segs[n_seg+1][1,3]-sec.segs[n_seg][1,3]), \
                        (sec.segs[n_seg+1][0,3]-sec.segs[n_seg][0,3]))
                        #phi=pi/4
                        sec.segs[n_seg][0:2,0:2]=array(\
                        [[cos(phi), -sin(phi)],
                         [sin(phi),  cos(phi)]])
                        #print '  seg',n_seg,'\n  ',sec.segs[n_seg]

                        #do last seg
                        #sec.segs[len(sec.segs)-1][0:2,0:2]=array(\
                        #[[cos(phi), -sin(phi)],
                        #[sin(phi),  cos(phi)]])
                    sec.segs[-1]=sec.tosecs_active[0].segs[0]
                    self.segs[sec.id]=sec.segs
                        #self.activate_sec(sec)
                
                        #print '     sec.segs=\n  ',sec.segs
                        #print '     self.segs[sec.id]=\n  ',self.segs[sec.id]
        
        #self.print_net(3)
        

    def rebuild_secpointer(self,secid):
        """
        Returns pointer to section.
        """
        # has it been replaced by elem id?
        if type(self.secs[secid])==types.StringType:
            # get track elem from id
            elem=self.sim.rebuild_pointer(self.secs[secid])
            #elem.rebuild_pointers()
            # now search section in section list
            sec_found=None
            for sec in elem.secs:
                if sec.id==secid:
                    sec_found=sec
                    # be sure that elem is relinked:
                    sec.elem=elem
                    break
                
            if not sec_found:
                print 'WARNING: netsim.rebuild_pointers '
                print '  unrecovered sec:',secid,'of',elem.id
                for sec in elem.secs:
                    print '    elem has sec',sec.id
            #else:
            #    'rebuild_secpointer',sec.id,elem.id
            return sec_found

        else:
            return self.secs[secid]
        
    def set_state(self,state_image):
        
        self.secs=state_image['secs']
        
        self.sec_id_current=state_image['sec_id_current']
        self.segs=state_image['segs']
        self.segs_total=state_image['segs_total']
        self.nodes_open=state_image['nodes_open']
        self.node_current=state_image['node_current']
        
        self.cars=state_image['cars']
        self.cars.ids_to_objs(self.sim)
        
        self.path=state_image['path']
        self.path_enter=state_image['path_enter']
        self.path_sec=state_image['path_sec']
        self.length=state_image['length']
        self.velocity=state_image['velocity']
        self.segs_ind=state_image['segs_ind']
        self.seg_length=state_image['seg_length']
        self.brake_factors=state_image.get('brake_factors',Array_db())
        self.brakedists=state_image.get('brakedists')


        
            
        # here introduce brake-times for older versions
        if state_image.has_key('times_brake'):
            self.times_brake=state_image.get('times_brake')
        else:
            self.times_brake=copy.deepcopy(self.velocity)
            self.times_brake.arrays=zeros(size(self.times_brake.arrays))
            

        
            
    def print_net (self,debug_level):
        if debug_level<=self.debug:
            print '----------------------- Net Topology ---------------------'
            print 'Sections:'
            for sec in self.secs.values():
                fromsec_ids=[]
                for fromsec in sec.fromsecs:
                    fromsec_ids.append(fromsec.id)
                tosec_ids=[]
                for tosec in sec.tosecs:
                    tosec_ids.append(tosec.id)
                print sec.id,': ',sec
                print fromsec_ids,' --> ',tosec_ids

            print 'Open nodes:'
            for node in self.nodes_open.values():
                print node.id,': ',node
                print node.localnode['direction'],\
                      ' at sec ',node.sec.id,\
                      ' xpos,ypos,zpos = ', node.pos.tolist()
        
            print '----------------------------------------------------------'

    def add_car(self,car,sec,dist):
        """
        Insert a carrier in the Netsim data structures on section
        sec with distance dist
        """
        if (sec.status=='active')&(len(sec.cars)<sec.car_max):
            car.sec=sec
            self.path.add(car.obj.type,array([0.0]))
        
            self.path_enter.add(car.obj.type,array([-dist]))
            self.length.add(car.obj.type,array([sec.length]))
        
            self.segs_total.add(car.obj.type,array([len(sec.segs)-1]) )
            self.seg_length.add(car.obj.type,
                                array([sec.length/(len(sec.segs)-1)]) )
            self.velocity.add(car.obj.type,array([0.0]))
            
            self.segs_ind.add(car.obj.type,array([self.segs.db[sec.id][0]]))
            self.cars.add(car.obj.type,[car])

            self.brake_factors.add(car.obj.type,
                                   [[ 1/(2*car.accel_comfort),
                                      1/(2*car.accel_emergency),
                                      1/(2*car.accel_failure)
                                      ]])
            self.times_brake.add(car.obj.type,[car.time_brake])
            
            # update Netsim's index arrays
            self.update_cars()

            #print 'Netsim: add_car ',car.id
            #print 'self.velocity.arrays',self.velocity.arrays
            #print 'self.path.db=\n',self.path.db
            #print 'self.path.arrays=\n',self.path.arrays
            #print 'self.car_ids.db',self.car_ids.db
            #print 'self.car_ids.objs',self.car_ids.objs
            #print 'sec.seg_ind= ',sec.seg_ind
            #print 'self.segs_ind.db=\n',self.segs_ind.db
            #print 'self.segs_ind.arrays=\n',self.segs_ind.arrays
       
            # make sure that Netsim is registered in maincycle now
            # ...as there is at least one car that needs to be moved
            self.sim.append_obj_iterate(self)

            # output only in gui mode
            #if self.sim.mode=='gui':
            #    self.sim.append_obj_output(self)
            
            return 0
        else:
            print 'WARNING: Netsim.add_car',car.id
            print ' Cannot place carrier because section is not active or free'
            return 1

    def set_car_params(self,car):
        """
        Set all necessary carrier parameters into already existing
        carrier in netsim's arrays
        """
        self.brake_factors.arrays[car.ind]=[ 1/(2*car.accel_comfort),
                                            1/(2*car.accel_emergency),
                                            1/(2*car.accel_failure)
                                            ]
        self.times_brake.arrays[car.ind]=car.time_brake
        
        
    def del_car(self,car):
        """
        Delete carrier from array structures.
        """
        #print 'self.cars: \n', self.cars.objs,self.cars.db

        if self.debug>=3:
            print 'Netsim.del_car: ',car.id,car.ind
        #car_ind=self.car_ids.get_index(car.id)
        self.path.del_index(car.ind)
        self.path_enter.del_index(car.ind)
        self.length.del_index(car.ind)
        self.segs_total.del_index(car.ind)
        self.seg_length.del_index(car.ind)
        self.velocity.del_index(car.ind)
        self.segs_ind.del_index(car.ind)
        self.cars.del_index(car.ind)
        self.brake_factors.del_index(car.ind)
        self.times_brake.del_index(car.ind)
        
        # special treatment for arrays that are not an Array_db
        inds=range(len(self.path_sec))
        inds.remove(car.ind)
        self.path_sec=take(self.path_sec,inds)
        
        self.update_cars()

        
            
        # make car leaving the section
        car.sec.leave(car)

        # decouple from all other carriers
        for dcar in self.cars.objs: dcar.control({'decouple':car})

            
        # if there are no more carriers left of this type
        #if not self.car_ids.db.has_key(car.obj.type):
        #    # delete this type also from the objects types database
        #    del self.carobjs[car.obj.type] 

        # delete from Netsim's carrier database
        #del self.cars[car.id]

        # the carrier sits no more on a section
        car.sec=None 

        # if there are no more cars to simulate, remove Netsim from maincycle
        #print 'self.cars: \n', self.cars.objs,self.cars.db
        if len(self.cars.objs)==0:
            #print 'remove Netsim objec from maincycle:',self.type
            self.sim.remove_obj_iterate(self)
            self.sim.remove_obj_output(self)
            
    def update_cars(self):
        '''Here the array index of each carrier (car.ind) is updated'''
        # update netsim's array index of all carriers
        ind=0
        for car in self.cars.objs:
            car.ind=ind
            ind=ind+1

        # update internal states that are a functions of previously
        # updated arrays.
        

        if self.debug>=5:
            print ' update the internal array index of all carriers'

        for type in self.cars.db.keys():
                self.sim.objs[type].update_arrays()

        for car in self.cars.objs:
            car.control()

        self.path_sec=self.path.arrays-self.path_enter.arrays
        #self.v=transpose([self.velocity.arrays])
        #print 'netsim:self.brake_factors=\n',self.brake_factors.arrays
        #print 'netsim:self.velocity.arrays=\n',self.velocity.arrays
        #print 'netsim:self.brake_factors.arrays=\n',self.brake_factors.arrays
        #print 'netsim:self.path.arrays=\n',self.path.arrays
        #print 'netsim:self.path_enter.arrays=\n',self.path_enter.arrays
        #print 'netsim:brakedists=\n',\
        #      self.brake_factors.arrays*self.v*self.v)
        
        #self.brakedists=self.v*self.v*self.brake_factors.arrays
        self.brakedists=transpose([self.velocity.arrays*self.velocity.arrays])\
                         *self.brake_factors.arrays+\
                         transpose([self.times_brake.arrays*\
                                    self.velocity.arrays])

        # update cars at netmon
        if self.netmon:
            self.netmon.update_cars()
            
    def add_sec(self,sec):
        """
        Insert new section into Netsim's data and array structure.
        """
        # secs database with all sections and id as key 
        sec.id=self.sec_id_current
        self.secs[sec.id]=sec
        self.sec_id_current=self.sec_id_current+1

        

    def add_elem(self, elem):

        for sec in elem.secs:
            
            # add section of element to Netsim's sections database 
            self.add_sec(sec)

            # determine track-element internal connections
            # and assign from and to sections
            fromnode=sec.fromnode['id']
            tonode=sec.tonode['id']
            sec.fromsecs=[]
            sec.tosecs=[]
            for sec_check in elem.secs:
                fromnode_check=sec_check.fromnode['id']
                tonode_check=sec_check.tonode['id']
                if fromnode==tonode_check:
                    sec.fromsecs.append(sec_check)
                if tonode==fromnode_check:
                    sec.tosecs.append(sec_check)

            # add open nodes to the open nodes directory
            if len(sec.fromsecs)==0:
                Node(self,sec.fromnode,sec)
                
            if len(sec.tosecs)==0:
                Node(self,sec.tonode,sec)
        self.update_elem(elem)
        
        
        
    def update_elem(self,elem):
        """ Updates geometrical changes to the track
        elem (no topological changes, as it does not make
        connections). Track element is replotted if it has no 'edit'
        status. For making connections use connect_elem method.  """

        for sec in elem.secs:
            sec.update()
            
        self.update_nodes(elem)        

        if self.netmon:
            if not(elem.status=='edit'):
                elem.plot()
            
        self.print_net(5)              
                        
    def isolate_elem(self,elem):
        """
        This function isolates a track element from the network by
        disconnecting it from all attached sections. In this way it is
        availlable for topological and geometrical editing.
        """
        for sec in elem.secs:
            if self.isolate_sec(sec):
                # reconnect sections in case the isolation failed
                self.connect_elem(elem)
                return 1

        if self.netmon:
            elem.plot()
        return 0

    def del_elem(self,elem):
        """
        Deletes element from network.
        """
        if self.isolate_elem(elem):
            return 1
        else:
            self.del_nodes(elem)

            # delete sections from data structures
            for sec in elem.secs:
                del self.secs[sec.id]
            elem.secs=[]
            return 0
        
        
    def activate_sec(self,sec):
        self.segs.add(sec.id,sec.segs)
        return 0
    
    def inactivate_sec(self,sec):

        occupied=0
        for car in self.cars.objs:
            if car.sec.id==sec.id:
                occupied=1
                break
            
        if not occupied:
            del self.segs[sec.id]

            # relink segment index of carrier arrays
            ind=0
            for car in self.cars.objs:
                self.segs_ind.arrays[ind]=self.segs.db[car.sec.id][0]
                ind=ind+1
        
            return 0
        else:
            return 1

    
        
    def isolate_sec(self,sec):

        track=sec.elem.obj # get track simulation object

        # get local nodes of track element
        fromnode=sec.fromnode['id']
        tonode =sec.tonode['id']
        
        contrack=None

        # fromnode to be disconnected must be
        # a connecting node (no internal node),
        # and it must be connected to another section
        # and the other section should not be active
        if ( track.nodes_connect.has_key(fromnode))&\
           ( len(sec.fromsecs)==1                 )&\
           ( len(sec.fromsecs_active)==0          ):

            # in this case, disconnect section input from attached section
            consec=sec.fromsecs[0]
            consec.tosecs=[]
            sec.fromsecs=[]

            # create new open nodes for the section and connected section
            Node(self,sec.fromnode,sec)
            contrack=consec.elem.obj
                       
            confromnode=consec.fromnode
            contonode=consec.tonode
            #print 'contonode: ',contonode
            Node(self,contonode,consec) #

            if self.netmon:
                consec.elem.plot()
                
        # the connected section is active ... isolation failed
        elif ( track.nodes_connect.has_key(fromnode))&\
             ( len(sec.fromsecs)==1                 )&\
             ( len(sec.fromsecs_active)>0           ):
            # cannot isolate from an active section
            return 1
            
        # tonode must be a connecting node (no internal node),
        # and it must be connected to another section
        # and the other section should not be active
        if ( track.nodes_connect.has_key(tonode))&\
           ( len(sec.tosecs)==1                 )&\
           ( len(sec.tosecs_active)==0          ):

            # in this case, disconnect section input from attached section
            consec=sec.tosecs[0]
            consec.fromsecs=[]
            sec.tosecs=[]
            
            # create new open nodes for the section and connected section
            Node(self,sec.tonode,sec)
            contrack=consec.elem.obj
            confromnode=consec.fromnode
            contonode= consec.tonode
            Node(self,confromnode,consec)

            if self.netmon:
                consec.elem.plot()

        # the connected section is active ... isolation failed
        elif ( track.nodes_connect.has_key(tonode))&\
             ( len(sec.tosecs)==1                 )&\
             ( len(sec.tosecs_active)>0           ):
            # cannot isolate from an active section
            return 1
        return 0   

    def connect_elem(self,elem):
        """
        Connexts track element with surrounding elements, when open nodes
        are close enough
        """

        for node in elem.nodes_open.values():
            node_closest=self.get_closest_node(node)
            if node_closest:
                if node.localnode['direction']=='in':
                    node.sec.fromsecs.append(node_closest.sec)
                    node_closest.sec.tosecs.append(node.sec)
                    
                elif node.localnode['direction']=='out':
                    node.sec.tosecs.append(node_closest.sec)
                    node_closest.sec.fromsecs.append(node.sec)

                # delete the two involved open nodes data...

                # ...from Netsims open node database
                del self.nodes_open[node.id]
                del self.nodes_open[node_closest.id]

                # ...from Track element's module data structure
                del node.sec.elem.nodes_open[node.localnode['id']]
                del node_closest.sec.elem.nodes_open[
                    node_closest.localnode['id']]
                
                # refresh also plot for attached track element
                if self.netmon:
                    node_closest.sec.elem.plot()

        
        if self.netmon:
            if elem.status!='edit':
                elem.plot()
        
        self.print_net(5)
               
    
    def add_node(self, node):
        
        node.id=self.node_current
        self.node_current=self.node_current+1
        
        #track=node.sec.elem.obj # get simulation object of the corresponding
        #                        # track element

        # add node to the open nodes database
        self.nodes_open[node.id]=node

        # and back-link also at the track element
        node.sec.elem.nodes_open[node.localnode['id']]=node

        
        
        return 0

    def del_nodes(self,elem):

        for node_local in elem.nodes_open.keys():
            node=elem.nodes_open[node_local]
            del self.nodes_open[node.id]
            del elem.nodes_open[node_local]
            
    def update_nodes(self,elem):
        """ updates the  position of all open nodes in a track element"""
                                     
        for node in elem.nodes_open.values():
            # get geometry and direction from local node
            # and pass it to global one in x,y,z array format
            node.pos=node.localnode['trans'][0:3,3]
           
    def get_closest_node(self, node_check):

        elem_check=node_check.sec.elem
        direction_check=node_check.localnode['direction']
        distance_min=10.0**64
        node_closest=None

        for node in self.nodes_open.values():
            elem=node.sec.elem
            distance=sqrt(sum((node.pos-node_check.pos)**2))

            if (distance < distance_min)&\
               (elem.id != elem_check.id)&\
               (node != node_check)&\
               (elem.status != 'edit')&\
               (node.localnode['direction'] != direction_check):
                distance_min=distance
                node_closest=node

        # is distance in the range of attraction ??
        if (distance_min<self.sim.node_attract):
            return node_closest
        else:
            return None
        
    
        
    def link(self):
        self.netmon=self.try_link_with('base.net.netmon')

        for node in self.nodes_open.values():node.netsim=self

        if self.sim.init_flag:
            #print 'link',self.id,':rebuild_pointers'
            #self.rebuild_pointers()
            if len(self.cars.objs)>0:
                self.sim.append_obj_iterate(self)
            # output only in gui mode
            #if self.sim.mode=='gui':
            #    self.sim.append_obj_output(self)

        
        
    def iterate(self):

        #self.v=transpose([self.velocity.arrays])
        self.brakedists=transpose([self.velocity.arrays*self.velocity.arrays])\
                         *self.brake_factors.arrays+\
                         transpose([self.times_brake.arrays*\
                                    self.velocity.arrays])
        
        # increases total path km
        self.path.arrays = self.path.arrays+\
                           self.velocity.arrays*self.sim.time_sample

        # determin path that each carrier has done in its section
        self.path_sec=self.path.arrays-self.path_enter.arrays
        #print 'netsim: self.velocity.arrays=\n',self.velocity.arrays
        #print '        brakedists=\n',self.brakedists
        #print '        self.path.arrays=\n',self.path.arrays
        #print '        self.path_sec=\n',self.path_sec
        #print '        self.path_sec.flat=\n',self.path_sec.flat
        #print '        self.length.arrays=\n',self.length.arrays
        
        # detect path of carriers that exeeds the present section length
        # and get the path index for those carriers
        self.sec_change_ind=nonzero(greater(self.path_sec,
                                            self.length.arrays))
        #ravel(segs[:,:2,3]).tolist()
        
        # assign enter path length, new section index and section length
        # to those carriers changing sections
        for ind in self.sec_change_ind:
            self.path_enter.arrays[ind]=self.path_enter.arrays[ind]+\
                                        self.length.arrays[ind]

            self.path_sec[ind]=self.path.arrays[ind]-\
                                self.path_enter.arrays[ind]
            
            # must be after path_sec after section change has been determined
            nextsec=self.cars.objs[ind].change_sec()
            self.length.arrays[ind]=nextsec.length
            self.segs_ind.arrays[ind]=self.segs.db[nextsec.id][0]
            self.segs_total.arrays[ind]=nextsec.segs_total
            self.seg_length.arrays[ind]=nextsec.seg_length

        
            
    
