#from classlib.track import *
from classlib.base import *
from classlib.man import *
from copy import copy
from RandomArray import *

class Transport(Task):
    """
    Task to transport a user to his/her destination
    """
    def init_task(self,user):
        """
        Choose a carrier management and create a task there to
        transport the user to its destination.
        """
        #print self.obj.id,'task',self.id,'init for',user['id']
        

        # store link to user
        self.res=user # define resource
        self.user=user # 
        self.dest_id=user['dest_id'] #save this for accounting after task
        
        # the global information is currently obtained from the
        # simulation environment (which will not exist in reality),
        # instead, a network management needs to be implemented...
        
        # go through all available simulation object types (=keys)
        for obj_type in self.obj.sim.objs.keys():
            if string.find(obj_type,'carman')!= -1: # is it a carman ?
                # start trans process for this carrier management
                self.add_proc(Trans_user,obj_type)
                
        # Now wait untill a process is completed

    def process_info(self,info):
        """
        Process newly received info from process.
        These may be new trip estimates. 
        """

        if info.has_key('trippath_est'):
            # return info to client process
            self.info.update(info)
            
    def update(self,proc):
        """
        Check whether process has been completed successfully,
        i.g if user arrived at its destination. If yes
        kill all other processes.
        I there are no more processes, the task is completed.
        """
        #print self.obj.id,'.Transport.update_procs for',self.user['id'],':',\
        #      self.id,proc.id,'\n  proc.status=',proc.status
        
        if proc.status=='done': # process has been successfully completed

            # compute real trip-time
            if self.info.has_key('triptime_start'):
                self.info['triptime']=self.obj.sim.time-\
                                       self.info['triptime_start']

            self.do_accounting()
            

            #print self.obj.id,':uuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu'
            #print self.user['id'],\
            #      'arrived at',self.dest_id

            #if self.info.has_key('trippath_est'):
            #    print '   Estimated trip length=',self.info['trippath_est']
            #if self.info.has_key('triptime_est'):
            #    print '   Estimated trip time=',self.info['triptime_est']
            #    print '   Trip time=',self.info['triptime']
                
        elif proc.status=='processing': # process started
            # this means a carrier of the corresponding carrier management,
            # has been assigned to the user, who is now boarding.
            
            # cancel all orders previously given to other carrier managements
            for proc_to_kill in self.procs.values():
                if proc_to_kill!=proc:
                    proc_to_kill.kill_proc()
            self.set_status('processing')
    
    def do_accounting_specific(self):
        """
        Do task specific accounting:
        Assign trip km and travel costs to specific user
        and do statistics for the user management.
        """
        #print '<<< do_accounting_specific',self.user['id'],'triptime_start=',self.info['triptime_start'],'time_start=',self.time_start

        # comments for future extensions:
        # doing accounting for users is a bit special because the
        # user needs to be charged (as a resource) and not the
        # client management (which is always the user management)

        # For the moment just count km, costs, etc for each user

        # update user data
        self.user['trips_total']+=1
        self.user['waitingtime_total']+=self.info['triptime_start']-\
                                         self.time_start
        self.user['trippath_total']+=self.info['trippath_est']
        self.user['trippath_est_total']+=self.info['trippath_est']
        self.user['triptime_total']+= self.info['triptime']
        #self.user['trippath_dev_total']+= (self.info['trippath']-\
        #                                   self.info['trippath_est'])**2
        self.user['triptime_var_total']+= (self.info['triptime']-\
                                           self.info['triptime_est'])**2
        
        self.user['triptime_relvar_total']+= \
               ((self.info['triptime']-self.info['triptime_est'])/\
                self.info['triptime'])**2
            
        # finally add costs which depend in general on the task
        # and therefore on user, total path, discounts...
        # The costs per km bases, which may depend on
        # the length of the trip is a method of the management 
        self.user['debits']+=self.obj.get_tripcosts(
            self.info['trippath_est'])
        
        

        
class Trans_user(Proc):
    """
    Process that orders a carrier management to transport the user
    in a carrier to its destination.
    """

    def init_proc(self,carman_type):

        """
        Make a transport order to this carrier management.
        """
        self.make_order(carman_type,'transport',
                        {'item_id':self.task.user['id'],
                         'dest_id':self.task.dest_id
                         })
        
        self.carman_type=carman_type # store management to deal with

    

    def process_info(self,info):
        """
        Process newly received info from resource.
        These may be new trip estimates. 
        """
        
        if info.has_key('trippath_est'):
            self.task.put_info(info)
            #print 'uuuuuuuuuuuuuuuuuuuu trippath_est=',info['trippath_est']
            #print 'uuuuuuuuuuuuuuuuuuuu triptime_est=',info['triptime_est']
            
        #print self.task.obj.id,'.Trans_user.process_info info status=',\
        #      self.status

        
    # used default...
    #def put_info(self,info):
    #    """
    #    Process information from carrier management
    #    about status of ordered task.
    #    """
       

class Experim(Man): 
    '''
    User management simulation object
    (Expetimental version)
    '''
    def init_obj_specific(self):
        

        self.users=self.resources

    

        # add operations that can be executed by an incoming message
        self.mes_ops['update_user']=self.update_user
        self.mes_ops['del_user']=self.del_user
        self.mes_ops['get_userdests']=self.get_userdests
        self.mes_ops['trans']=self.trans
        #self.mes_ops['put_dest']=self.put_dest

        self.mes_ops['unknown_dest']=self.unknown_dest

        # The table task_classes is mapping task names to task classes.
        self.task_classes['transport']=Transport

        # initialise destinations, mapping destination id to track man id
        self.dests={}
        
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        #self.poll_dests_task_id='' # will be done when first user is added
        #                           # or during set_state with saved data
        #self.poll_dests_task_id=self.make_order('poll_destinations')
        self.poll_dests_task_id=''

    def init_costs_specific(self):
        """
        Management specific cost initialization.
        To be overwritten.
        """
        self.add_cost_item('revenue',
                           {'name':'Revenue',
                            'value':0.0,
                            'mode':'read'},# determined in do_costs_current
                           )
        self.add_cost_item('revenue_per_h',
                           {'name':'Revenue per hour',
                            'value':0.0,
                            'currency':self.sim.currency+'/h',
                            'mode':'read'},# determined in do_costs_current
                           )
        self.add_cost_item('costs_trip_fixed',
                           {'name':'Fixed trip costs',
                            'value':1.0,
                            'currency':self.sim.currency,
                            'mode':'write'}
                           )
        self.add_cost_item('costs_trip_per_distance',
                           {'name':'Costs per distance',
                            'value':0.2,
                            'currency':self.sim.currency+'/km',
                            'mode':'write'}
                           )
        

            
    def get_state(self):
        '''Prepare state image for save operation'''
        

        state_image=self.get_state_man({})
        #state_image['users']=self.users
        state_image['dests']=self.dests
        
        state_image['poll_dests_task_id']=self.poll_dests_task_id
        #state_image['costs_per_dist']=self.costs_per_dist
        #state_image['']=self.
        #state_image['']=self.
        
        return state_image
        
    def set_state(self,state_image):
        '''Set state while opening simulation file'''
        
        #print 'set_state',self.id
        self.set_state_man(state_image)
        self.users=self.resources
        self.dests=state_image['dests']

        # get id from poll_dests_task
        if state_image.has_key('poll_dests_task_id'):
            #print ' has poll_dests_task_id:',state_image['poll_dests_task_id']
            self.poll_dests_task_id=state_image['poll_dests_task_id']
        else:
            self.poll_dests_task_id=''
            #self.poll_dests_task_id=self.make_order('poll_destinations')
        

        #self.=state_image['']
        #self.=state_image['']
        #self.=state_image['']

    def rebuild_pointers(self):
        """
        Rebuild pointers that have been removed before saving.
        This function is called AFTER all objects have been
        initialized and cross-linked.
        """
        # get id from poll_dests_task
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        if self.poll_dests_task_id=='':
            #print '  make poll_dests_task'
            self.poll_dests_task_id=self.make_order('poll_destinations')
        else:
            #print '  has poll_dests_task procs?',\
            #      len(self.tasks[self.poll_dests_task_id].procs)
            if len(self.tasks[self.poll_dests_task_id].procs)==0:
                self.tasks[self.poll_dests_task_id].kill()
                self.poll_dests_task_id=self.make_order(\
                        'poll_destinations')
        
    #def start_poll_destinations(self):
    #    """
    #    Start the poll destination task, if not alredy started...
    #    """
    #    if self.poll_dests_task_id=='':
    #        self.poll_dests_task_id=self.make_order('poll_destinations')

    def get_resource_name(self):
        """
        Returns string with name of resource that is managed by this management
        To be overwritten.
        """
        return 'User'
    
    def reset_statistics_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular type.

        To be overwritten.
        """
        # do through resources here as this function may be called
        # from the general management class
        #print 'reset_statistics_specific for',self.id
        for user in self.resources.values():
            self.reset_statistics_user(user)

    def reset_statistics_user(self,user):
        """
        Initialize and reset all statistical data for  this user. 
        """
        #print 'reset_statistics_user for ',user['id']
        user['trips_total']=0
        user['waitingtime_total']=0.0
        user['trippath_total']=0.0
        user['trippath_est_total']=0.0
        user['triptime_est_total']=0.0
        user['triptime_total']=0.0
        #user['trippath_dev_total']=0.0
        user['triptime_var_total']= 0.0
        user['triptime_relvar_total']=0.0
            
        
    def format_statistics_specific(self,descriptor,dataset):
        """
        Format management specific contents for statistical data.
        """
        descriptor_specific=[
            {'key':'Total # of trips',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.2,
             },
            {'key':'Total distance traveled',
             'type':'real',
             'printformat':'%9.1f',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.4,
             'units':{'km':1}
            },
            {'key':'Average velocity',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.4,
             'units':{'km/h':1}
             },
            {'key':'Average waiting times',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**5},
             'mode':'read',
             'width':0.4,
             'units':{'s':1}
             },
            {'key':'Average delays',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**5},
             'mode':'read',
             'width':0.4,
             'units':{'s':1}
             },
            {'key':'Average relative delays',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**2},
             'mode':'read',
             'width':0.4,
             'units':{'%':1}
             }
            ]

        descriptor+=descriptor_specific

        n=0
        trips_sum=0
        distance_sum=0.0
        velocity_sum=0.0
        waitingtime_sum=0.0
        delaytime_sum=0.0
        reldelay_sum=0.0

        # calculating statistics over all users
        for user in self.users.values():
            if (user['trippath_total']>0.0)&\
                   (user['triptime_total']>0.0)&\
                   (user['trips_total']>0):
                n+=1
                trips_sum+=user['trips_total']
                distance_sum+=user['trippath_total']/1000
                velocity_sum+=user['trippath_total']/ \
                                  user['triptime_total']*3.6
                waitingtime_sum+=user['waitingtime_total']/\
                                  user['trips_total']
                delaytime_sum+=sqrt(user['triptime_var_total']/\
                                    user['trips_total'])
                reldelay_sum+=sqrt(user['triptime_relvar_total']/\
                                 user['trips_total'])*100
          
        if n>0:
            dataset['Total # of trips']=trips_sum
            dataset['Total distance traveled']=distance_sum
            dataset['Average velocity']=velocity_sum/n
            dataset['Average waiting times']=waitingtime_sum/n
            dataset['Average delays']=delaytime_sum/n
            dataset['Average relative delays']=reldelay_sum/n
        else:
            dataset['Total # of trips']=0
            dataset['Total distance traveled']=0.0
            dataset['Average velocity']=0.0
            dataset['Average waiting times']=0.0
            dataset['Average delays']=0.0
            dataset['Average relative delays']=0.0
        
    def do_costs_current_specific(self):
        """
        Calculates current costs and modifies the costs directory
        """
        # add up debits of all users
        revenue=0.0
        for user in self.users.values():
            if user.has_key('debits'):
                revenue+=user['debits']
                
        self.set_cost_value('revenue',revenue)
        if self.sim.get_duration('h')<1.0:
            self.set_cost_value('revenue_per_h',0.0)
        else:
            self.set_cost_value('revenue_per_h',revenue/\
                                self.sim.get_duration('h'))
        
    
       
    
    def get_tripcosts(self,path=1000.0):
        """
        Return the costs for a given path.
        When called without arguments, it will return the costs par km.
        """
        return self.get_cost_value('costs_trip_fixed')+\
               0.001*path*self.get_cost_value('costs_trip_per_distance')

    def reset_costs_specific(self):
        """
        Sets all management specific
        costs in the costs directory that occurred during
        the current simulation to zero.

        
        """
        # this may go to man class, using reset_costs_resource
        for user in self.resources.values():
            self.reset_costs_user(user)
    
    def reset_costs_user(self,user):
        """
        Initialize and reset all cost data for  this user. 
        """
        user['debits']=0.0

    def format_resources_specific(self,header,table):
        """
        Format table with management specific contents for resource data.
        To be overwritten.
        """
        header.append({'key':'trips',
                       'type':'int',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.2,
                       })
        header.append({'key':'traveled',
                       'type':'real',
                       'printformat':'%.1f',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.4,
                       'units':{'km':1}
                       })

        header.append({'key':'speed av.',
                       'type':'real',
                       'printformat':'%.2f',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.4,
                       'units':{'km/h':1}
                       })
        header.append({'key':'waited av.',
                       'type':'real',
                       'printformat':'%.1f',
                       'valid':{'min':0,'max':10**4},
                       'mode':'read',
                       'width':0.4,
                       'units':{'s':1}
                       })
        
        header.append({'key':'delay av.',
                       'type':'real',
                       'printformat':'%.1f',
                       'valid':{'min':0,'max':10**5},
                       'mode':'read',
                       'width':0.4,
                       'units':{'s':1}
                       })
        #header.append({'key':'rel. delay av.',
        #               'type':'real',
        #               'printformat':'%.2f',
        #               'valid':{'min':0,'max':10**2},
        #               'mode':'read',
        #               'width':0.4,
        #               'units':{'%':1}
        #               })
        
        header.append({'key':'debits',
                       'type':'real',
                       'printformat':'%.2f',
                       'width':0.4,
                       'mode':'read',
                       'valid':{'min':0,'max':10**8},
                       'units':{self.sim.currency:1},
                       })
        
            
        for row in table:
            user=self.users[row['id']]
            if (user['trippath_total']>0.0)&\
                   (user['triptime_total']>0.0)&\
                   (user['trips_total']>0):

                row['trips']=user['trips_total']
                row['traveled']=user['trippath_total']/1000
                row['speed av.']=user['trippath_total']/ \
                                  user['triptime_total']*3.6
                row['waited av.']=user['waitingtime_total']/\
                                  user['trips_total']
                row['delay av.']=sqrt(user['triptime_var_total']/\
                                  user['trips_total'])
                #row['rel. delay av.']=sqrt(user['triptime_relvar_total']/\
                #                  user['trips_total'])*100
                #row['debits']=user['debits']
                

   

    def register_specific(self,resource,message={}):
        """
        Management specific resource registration.
        Will be initialization of statistical and cost data.
        """

        # print 'register_specific:',resource['id']
        
        # this may move to general register function in man
        self.reset_statistics_user(resource)
        self.reset_costs_user(resource)
        
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        # already done in set_state
        #if self.poll_dests_task_id:
        #    self.tasks[self.poll_dests_task_id].kill()
        #    #self.poll_dests_task_id=self.make_order('poll_destinations')
        
    def update_user(self,message):
        user_id=message['from']
        #print 'update_user******************************************'
        #if not self.users.has_key(user_id):
        #    # add new user to users db
        #    self.users[user_id]={'id':user_id,
         #                        'status':'edit',
        #                         'dest_id':''}
        #    user=self.users[user_id]
        #    
        #    # initialize statistics and cost data
        #    self.reset_statistics_user(user)
        #    self.reset_costs_user(user)

        user=self.users[user_id]
        #print '   resources=',self.resources
        #print '   user:users=',self.users
        if self.debug>=5:
            print self.type,'.update_user:',message['from'],\
                  user['status'],'-->',message['status'],\
                  'dest_id=',user['dest_id']

        if user['status']!=message['status']: # check if user status changed
            # user went into reside status ?
            if message['status']=='reside':
                # check if user arrived at the destired destination
                #if user['status']=='traveling':
                    #print 'New location_id:',message['location_id']
                    #if user['dest_id']==message['location_id']:
                    #    print 'Arrived:',user['id'],\
                    #          'with destination ',user['dest_id'],\
                    #          'completed at',message['location_id']

                    
                # store new status and user info in local user db
                self.set_user_status(user,message)

                
            else:
                # store new status and user info in local user db
                self.set_user_status(user,message)
                
    def set_user_status(self,user,message):
        '''
        Reads user info from message and and inserts it into user
        database.
        '''
        user['status']=message['status']
        if message.has_key('location_id'):
            user['location_id']=message['location_id']
            
    def del_user(self,message):
        user_id=message['from']
        if self.users.has_key(user_id):
            del self.users[user_id]

            if len(self.users)==0:
                self.sim.remove_obj(self)

    def get_userdests(self,message):
        """
        Send a message back to the user that contains all destinations
        that the user can currently choose from.
        """
        self.send_mes({'to':message['from'],
                       'op':'choose_dest',
                       #'n_dests':len(self.dests),
                       'dest_ids':self.dests.keys()
                       })
        
    def trans(self,message):
        """
        User wants to be transported to a new destination
        """
        #print '>>>',self.id,'trans user',message['from'],'at',self.sim.time
        
        user=self.users[message['from']]
        user['dest_id']=message['dest_id']

        # create a task...
        
        self.make_order('transport',user)
    

    #def charge_user(self,task):
    #    """
    #    Charge user for a the given transport task.
    #    """
     
            
    def unknown_dest(self,message):
        
        ''' User encountered an unknown destination.  Probably the
        current destination table is outdated. Poll the track management
        of the old user destination for a new destination table.
        User with invalid detination will be branched to a stop,
        where it will receive a new destination.
        '''
        print 'WARNING:',self.users[message['from']],\
              'encoutered unknown destination'
        
        #user=self.users[message['from']]
        #if user.has_key('dest_id'):
        #    if self.dest.has_key(user['dest_id']): # key is trackman id
        #        self.make_order('poll_destinations')
        #        #self.poll_dest(self.dest[user['dest_id']])
                
    
                        
    def output(self):
        """
        This function is called each dot seconds to make regular updates.
        """
        #print self.id,'output*******************'
        #print 'Userman dest????????????',len(self.dests)
            
            
        #if len(self.dest)==0: # no destinations available
        #    self.make_order('poll_destinations')
            #print ' => poll all track managements for available destinations'
            #self.poll_dest()

            
        # read and process all messages in the message buffer
        self.read_mes()

    
