
"""
This script is based on the script_users file, and provides
also logging capability of number of waiting users at stops.

This script will open a predefined its simulation file,
it will add a specified number of vehicles and users,
it will run the simulation for a specified amout of time,
it will safe simulation and export results, if desired.

The simulation will run for various numbers of users specified in the
loop declarator and for each number of users a result-simulation file
will be produced.  But feel free to change any parameter you like or
change two parameters in a double loop.

"""

# defines file path relative to the position of this script file
# Note: under windows use double-backslash (\\) to separate directories. 
simdir='projects'

# the simulation file with the (empty) network.
simfile='fiera_track'

# load the simulator package into memory
import its

# import OS independent directory path operations
from os import path

for n_users in [200,400,500]:
    
    # initialize the simulator with the simulation file
    sim=its.Sim( path.join(simdir,simfile+'.its') )
    
    # make sure the track is active
    sim.activate_all()
    
    # initialize the vehicle simulation object of the desired carrier type
    # you can skip this part if vehicles are already present in the simulation
    object_vehicle=sim.append_obj('test.car.experim')
    
    # this instruction adds an equal number of vehicle-modules
    # to each stop of the network, using the given vehicle parameters
    object_vehicle.multi_add(number=200,
                             zone='everywhere',
                             length=3.0,
                             accel_comfort=2.5,
                             accel_emergency=5.0,
                             accel_failure=10.0,
                             time_brake=0.1)

    # initialize the user simulation object
    object_user=sim.append_obj('base.user.test_driver')

    # and add the desired number of users
    object_user.multi_add(number=n_users,
                          zone='everywhere',
                          time_boarding=7,
                          time_unboarding=7,
                          velocity_walk=0.5,
                          )
                          
    # init log-file:
    simfile_log=simfile+'_users'+str(n_users)+'_log.txt'  
    sim.init_log_waiting_users(simfile_log)

    # start the simulation with the end-of-simulationtime in sec as parameter
    sim.start(time_end=600)

    # compose name of output simulation file
    simfile_output=simfile+'_users'+str(n_users)

    # save simulation 
    print 'Saving',simfile_output+'.its'
    sim.save( path.join(simdir,simfile_output+'.its') )
        
    

    # if you also want a report written and saved for this simulation add...
    sim.export_results(path.join(simdir,simfile_output+'.txt'))

    # before running the simulation with the next set of parameters,
    # delete the old simulation in order to safe memory
    del sim
